# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from colorpalette import Color

__all__ = ('Handle', 
           'MassStreamHandle', 'MolarStreamHandle', 'VolumetricStreamHandle', 
           'CapitalNodeHandle', 'CarbonColorHandle', 'CarbonHandle', 
           'StreamNode', 'FeedNode', 
           'ProductNode', 'ProcessNode', 
           'sankey_figure', 'sankey_data',
           'node_dict', 'link_dict')

cycle_colors = ('aliceblue', 'chocolate', 'peachpuff', 
                'palevioletred', 'orange', 'springgreen', 'slategrey', 'goldenrod', 
                'mintcream', 
                'azure', 'darkseagreen', 'chartreuse', 'darkgray', 'khaki', 
                'plum', 'floralwhite', 'darkturquoise', 'steelblue', 'lightcyan', 
                'mistyrose', 'lightgray', 'lime', 'forestgreen',
                'darkblue', 'lightslategrey', 'burlywood', 'lightskyblue',
                'lightgrey', 'palegreen', 'indianred', 'royalblue', 
                'darkslateblue', 'blanchedalmond', 'seashell', 'dimgrey',
                'orchid', 'lightsalmon', 'beige', 'seagreen', 'teal',
                'lightblue', 'yellow', 'cornflowerblue', 'rosybrown', 'orangered',
                'peru', 'mediumorchid', 'mediumslateblue', 'darkgrey', 
                'slateblue', 'purple', 'lightslategray', 'limegreen', 'olivedrab', 
                'lightpink', 'olive', 'lightgoldenrodyellow', 'dimgray', 'mediumblue',
                'greenyellow', 'cyan', 'skyblue', 'green', 'ghostwhite',
                'hotpink', 'mediumvioletred', 
                'lawngreen', 'turquoise', 'powderblue', 'navy', 'gray', 
                'deeppink', 'lightcoral', 'mediumseagreen', 'maroon', 'honeydew', 
                'lavender', 'mediumturquoise', 'darkcyan', 'darksalmon',
                'paleturquoise', 'tomato', 'darkkhaki', 'darkgreen',
                'firebrick', 'ivory', 'mediumspringgreen', 'oldlace',
                'papayawhip', 'lemonchiffon', 'lightyellow', 'aquamarine', 
                'mediumpurple', 'aqua', 'blue', 'salmon', 'blueviolet',
                'lightseagreen', 'whitesmoke', 'linen', 'mediumaquamarine',
                'rebeccapurple', 'deepskyblue', 'sienna', 'violet', 'black',
                'darkgoldenrod', 'darkorchid', 'yellowgreen', 'darkviolet', 
                'pink', 'slategray', 'magenta', 'gainsboro', 'wheat', 'dodgerblue', 
                'fuchsia', 'cornsilk', 'palegoldenrod', 'saddlebrown', 
                'darkslategrey', 'indigo', 'snow', 'darkslategray', 
                'red', 'gold', 'coral', 'bisque', 'midnightblue', 
                'navajowhite', 'tan', 'moccasin', 'silver', 'brown', 
                'darkorange', 'darkred', 'antiquewhite', 'grey', 'crimson', 
                'white', 'sandybrown', 'darkmagenta', 'lavenderblush', 
                'lightgreen', 'thistle', 'darkolivegreen', 'lightsteelblue',
                'cadetblue')

# %% Helpful functions

def stream_cost(stream):
    try:
        return abs(stream.cost)
    except:
        return 0. # Doesn't matter, this is used only for sorting

def abbreviate_word(word):
    if len(word) < 5: return word 
    original_word = word
    word = word.lower()
    vocals = 'aeiouy'
    for index, letter in enumerate(word):
        if letter in vocals: break
    for index, letter in enumerate(word[index:], index + 1):
        if letter not in vocals: break
    for index, letter in enumerate(word[index:], index + 1):
        if letter in vocals: 
            index -= 1
            break
    if index < len(word):
        return original_word[:index] + '.'
    else:
        return original_word
            
def abbreviate_name(name):
    words = name.split(' ')
    return ' '.join([abbreviate_word(i) for i in words])


def get_unit_group(stream, unit_groups):
    for group in unit_groups:        
        if stream in bst.utils.streams_from_units(group.units): return group
        
def reduced_streams(streams, unit_groups):
    same_IDs = {}
    for s in streams:
        ID = s.ID
        if ID in same_IDs:
            originals = same_IDs[ID]
            append = True
            unit_group = get_unit_group(s, unit_groups)
            for original in originals:
                if get_unit_group(original, unit_groups).name == unit_group.name:
                    original.mix_from([original, s])
                    append = False
                    break
            if append: originals.append(s) 
        else:
            same_IDs[ID] = [s]
    return sum(same_IDs.values(), [])


# %% Main class used to handle the creating of Sankey plots

class Handle: # pragma: no coverage
    """
    Create a Handle object for creating Sankey diagrams. You can subclass and
    implement the following to customize streams and processes.
    
    **Abstract methods**
    
    stream_width(stream) -> float
        Return link width.
    stream_color(stream) -> tuple[float, float, float]
        Return link color in RGB. Defaults to gray.
    process_color(unit, index) -> tuple[float, float, float]
        Return node color in RGB. By default, the color of proces nodes cycle 
        through colors defined in the variable, `biosteam.plots.sankey.color_cycle`.
    
    """
    __slots__ = (
        'size', 'nodes_index',
        'main_feeds', 'max_feeds',
        'main_products', 'max_products',
        'ignore',
    )
    
    def __init__(self, main_feeds=None, main_products=None, 
                 max_feeds=3, max_products=3, ignore=None, 
                 **kwargs):
        self.nodes_index = {} #: Dict[Object, ProcessNode] Object - node pairs.
        self.size = 0 #: [int] Number of process nodes created.
        self.main_feeds = main_feeds
        self.main_products = main_products
        self.max_feeds = max_feeds
        self.max_products = max_products
        self.ignore = ignore or ()
        self._init(**kwargs)
    
    def _init(self): pass
    
    def filter_streams(self, streams):
        has_width = self.stream_width
        return [i for i in streams if has_width(i)]
    
    stream_width = bst.utils.AbstractMethod
    stream_color = bst.utils.AbstractMethod
    process_color = bst.utils.AbstractMethod
    
    def next_index(self):
        index = self.size
        self.size = index + 1
        return index
    
    def feed_node(self, feed, name=None):
        feed_node = FeedNode(self, feed, self.next_index(), name)
        self.nodes_index[feed] = feed_node
        return feed_node
    
    def product_node(self, product, name=None):
        product_node = ProductNode(self, product, self.next_index(), name)
        self.nodes_index[product] = product_node
        return product_node
    
    def process_node(self, unit_group):
        units = unit_group.units
        process_node = ProcessNode(self, unit_group.name, self.next_index(), units)
        for i in units: self.nodes_index[i] = process_node 
        return process_node
    
    def nodes(self, unit_groups):
        if isinstance(unit_groups, dict):
            unit_groups = [bst.UnitGroup(i, j) for i,j in unit_groups.items()]
        nodes = []
        streams = set()
        streams_from_units = bst.utils.streams_from_units
        for group in unit_groups:
            nodes.append(self.process_node(group))
            streams.update(streams_from_units(group.units))
        streams = self.filter_streams(streams)
        nodes_index = self.nodes_index
        feeds = [i for i in streams if i._source not in nodes_index and i._sink and i not in self.ignore and not i.isempty()]
        main_feeds = self.main_feeds or sorted(feeds, key=stream_cost, reverse=True)[:self.max_feeds]
        feeds = reduced_streams(main_feeds, unit_groups)
        for feed in feeds:
            node = self.feed_node(feed)
            nodes.append(node)
        products = [i for i in streams if i._sink not in nodes_index and i._source and i not in self.ignore and not i.isempty()]
        main_products = self.main_products or sorted(products, key=stream_cost, reverse=True)[:self.max_products]
        products = reduced_streams(main_products, unit_groups)
        for product in products:
            node = self.product_node(product)
            nodes.append(node)
        return nodes
    
    def sankey_data(self, unit_groups, *args, **kwargs):
        return sankey_data(self.nodes(unit_groups), *args, **kwargs)
    
    def sankey_figure(self, data, *args, **kwargs):
        return sankey_figure(data, *args, **kwargs)

# %% Custom hanldes

class CapitalNodeHandle(Handle): # pragma: no coverage
    """
    Create a CapitalNodeHandle object that represents node colors by installed 
    equipment cost of the process.
    
    """
    __slots__ = ('max_installed_cost', 'process_color_source', 'shade')
    
    def _init(self, max_installed_cost=None, process_color_source=None, shade=0.75):
        self.max_installed_cost = max_installed_cost
        self.process_color_source = process_color_source or bst.utils.colors.CABBI_orange
        self.shade = shade

    def installed_cost_color(self, installed_cost):
        scale = 100 * self.shade * installed_cost / self.max_installed_cost
        return self.process_color_source.shade(scale)

    def process_color(self, units, index):
        if self.max_installed_cost:
            installed_cost = sum([i.installed_cost for i in units])
            RGB = self.installed_cost_color(installed_cost).RGB
            return "rgba(%d, %d, %d, 1.0)" %tuple(RGB)
        else:
            return cycle_colors[index % len(cycle_colors)]
        
    def process_colorbar(self, N_levels=25, orientation='vertical'):
        colors = [self.installed_cost_color(0.).RGBn,
                  self.installed_cost_color(self.max_installed_cost).RGBn]
        return bst.plots.color_bar(colors, label='Installed equipment cost [million USD]',
                                   vmin=0, vmax=self.max_installed_cost / 1e6,
                                   N_levels=25, orientation=orientation)

class MassStreamHandle(Handle): # pragma: no coverage
    """
    Create a MassHandle object that represents stream widths by mass flow.
    
    """
    __slots__ = ()
    def stream_width(self, stream): return stream.F_mass


class MolarStreamHandle(Handle): # pragma: no coverage
    """
    Create a MolarHandle object that represents stream widths by molar flow.
    
    """
    __slots__ = ()
    def stream_width(self, stream): return stream.F_mol


class VolumetricStreamHandle(Handle): # pragma: no coverage
    """
    Create a VolumetricHandle object that represents stream widths by volumetric 
    flow.
    
    """
    __slots__ = ()
    def stream_width(self, stream): return stream.F_vol


class CarbonStreamHandle(Handle): # pragma: no coverage
    """
    Create a CarbonStreamHandle object that represents stream widths by
    carbon flow by weight.
    
    """
    __slots__ = ()
    def stream_width(self, stream): return stream.get_atomic_flow('C')


class CarbonColorHandle(Handle): # pragma: no coverage
    """
    Create a CarbonColorHandle object that represents stream color by
    carbon content by weight.
    
    """
    __slots__ = ()
    
    def stream_carbon_content(self, stream):
        return stream.get_atomic_flow('C') * 12.01 / stream.F_mass
       
    def carbon_content_color(self, carbon_content):
        scale = 95. * (1 - carbon_content)
        return bst.utils.colors.grey.shade(10).tint(scale)
    
    def stream_color(self, stream):
        carbon_content = self.stream_carbon_content(stream)
        RGB = self.carbon_content_color(carbon_content).RGB
        return "rgba(%d, %d, %d, 1.0)" %tuple(RGB)

    def stream_colorbar(self, N_levels=25, orientation='vertical'):
        colors = [self.carbon_content_color(0.).RGBn,
                  self.carbon_content_color(1.).RGBn]
        return bst.plots.color_bar(colors, label='Carbon content [wt. %]',
                                   vmin=0, vmax=100, N_levels=25,
                                   orientation=orientation)


class CarbonHandle(CarbonColorHandle, MassStreamHandle, CapitalNodeHandle): # pragma: no coverage
    __slots__ = ()


# %% Main classes used by Handle objects to create Sankey plots

class StreamNode: # pragma: no coverage
    """
    Abstract class for stream nodes.
    
    """
    __slots__ = ('handle', 'name', 'stream', 'index')
    
    def __init__(self, handle, stream, index, name=None):
        self.handle = handle
        self.name = name or stream.ID.replace('_', ' ')
        self.index = index
        self.stream = stream

    def color(self):
        return self.handle.stream_color(self.stream)
    
    def __repr__(self):
        return f"<{type(self).__name__}: {self.name}>"


class FeedNode(StreamNode): # pragma: no coverage
    """
    Create a FeedNode object that represents a feed in a Sankey diagram.
    
    """
    __slots__ = ()
    
    def links(self):
        stream = self.stream
        sanky_group = self.handle.nodes_index[stream.sink]
        return [Link(self.handle, self, sanky_group, [stream])]
    
    
class ProductNode(StreamNode): # pragma: no coverage
    """
    Create a ProductNode object that represents a product in a Sankey diagram.
    
    """
    __slots__ = ()
    
    def links(self):
        stream = self.stream
        sanky_group = self.handle.nodes_index[stream.source]
        return [Link(self.handle, sanky_group, self, [stream])]
    

class ProcessNode: # pragma: no coverage
    """
    Create a ProcessNode object that represents a process in a Sankey diagram.
    
    """
    __slots__ = ('handle', 'name', 'index', 'units')
    
    def __init__(self, handle, name, index, units):
        self.handle = handle
        self.name = name
        self.index = index
        self.units = units
        
    def links(self):
        handle = self.handle
        nodes_index = handle.nodes_index
        units = frozenset(self.units)
        streams = bst.utils.outlets(units)
        streams = handle.filter_streams(streams)
        process_streams = [i for i in streams if i.sink not in units and i.sink in nodes_index]
        all_process_nodes = {nodes_index[i.sink] for i in process_streams}
        streams_by_process_node = {i:[] for i in all_process_nodes}
        for i in process_streams:
            process_node = nodes_index[i.sink]
            streams_by_process_node[process_node].append(i)
        return [Link(handle, self, sink, streams) for sink, streams in streams_by_process_node.items()]
  
    def color(self):
        return self.handle.process_color(self.units, self.index)

    __repr__ = StreamNode.__repr__


class Link: # pragma: no coverage
    """
    Create a Link object that represents a link in a Sankey diagram.
    
    """
    __slots__ = ('handle', 'source', 'sink', 'streams')
    
    def __init__(self, handle, source, sink, streams):
        self.handle = handle #: [Handle]
        self.source = source #: [ProcessNode or StreamNode]
        self.sink = sink #: [ProcessNode or StreamNode]
        self.streams = streams #: [Stream]

    @property
    def stream(self):
        streams = self.streams
        if len(streams) == 1: return streams[0]
        return bst.Stream.sum(streams)

    def value(self):
        return self.handle.stream_width(self.stream)
    
    def color(self):
        return self.handle.stream_color(self.stream)
        
    def __repr__(self):
        return f"<{type(self).__name__}: {self.source.name} - {self.sink.name}>"


# %% Functions that use plotly to create diagrams

def sankey_figure(data, **kwargs): # pragma: no coverage
    import plotly.graph_objects as go
    return go.Figure(data=data, **kwargs)

def sankey_data(nodes, arrangement=None, 
                node_kwargs=None, link_kwargs=None,
                links=None, orientation=None, add_label=None): # pragma: no coverage
    import plotly.graph_objects as go
    node_kwargs = node_kwargs or {}
    link_kwargs = link_kwargs or {}
    # if arrangement is None: arrangement = 'snap'
    if links is None: links = sum([i.links() for i in nodes], [])
    return go.Sankey(
        orientation=orientation,
        arrangement = arrangement,
        node = node_dict(nodes, add_label, **node_kwargs),
        link = link_dict(links, **link_kwargs),
    )

def node_dict(nodes, add_label=None, **kwargs): # pragma: no coverage
    nodes = sorted(nodes, key=lambda x: x.index)
    dct = {
        'color': [i.color() for i in nodes],
        **kwargs
    }
    if add_label is None: add_label = True
    if add_label: dct['label'] = [i.name for i in nodes]
    return dct
                                             
def link_dict(links, **kwargs): # pragma: no coverage
    return {
        'source': [i.source.index for i in links],
        'target': [i.sink.index for i in links],
        'value': [i.value() for i in links],
        'color': [i.color() for i in links],
        **kwargs
    }



# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2021-2024, Yalin Li <mailto.yalin.li@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

import math, flexsolve as flx
import biosteam as bst
from warnings import warn

__all__ = ('SludgeHandling', 'BeltThickener', 'SludgeCentrifuge')


# %%

class SludgeHandling(bst.Unit):
    """
    A generic class for handling of wastewater treatment sludge [6]_. Two pumps 
    (one for the supernatant and one for sludge) are included.

    Separation split is determined by the moisture (i.e., water) content
    of the sludge, soluble chemicals will have the same split as water,
    insolubles chemicals will all go to the retentate.

    Parameters
    ----------
    ins : 
        Influent
    outs : 
        * [0] water-rich supernatant
        * [1] solid-rich sludge
    sludge_moisture : float
        Moisture content of the sludge, [wt. % water].
    solubles : tuple
        IDs of the soluble chemicals.
        Note that all chemicals that are not included in this tuple and not
        locked as gas phase (i.e., `chemical.locked_state` != "g") will be
        treated as solids in simulation.

    """

    SKIPPED = False
    _graphics = bst.Splitter._graphics
    _ins_size_is_fixed = False
    _N_outs = 2
    auxiliary_unit_names = ('effluent_pump', 'sludge_pump')


    def _init(self, sludge_moisture=0.96, solubles=()):
        self.sludge_moisture = sludge_moisture
        self.solubles = tuple(solubles)
        self.solids = tuple(i.ID for i in self.chemicals
                            if (i.ID not in solubles) and (i.locked_state!='g'))
        ID = self.ID
        self._mixed = bst.Stream(f'{ID}_mixed')
        # Add '.' in ID for auxiliary units
        self.effluent_pump = bst.Pump(f'.{ID}_eff_pump', ins=self.outs[0].proxy(f'{ID}_eff'))
        self.sludge_pump = bst.Pump(f'.{ID}_sludge_pump', ins=self.outs[1].proxy(f'{ID}_sludge'))


    @staticmethod
    def _mc_at_split(split, solubles, mixed, eff, sludge, target_mc):
        eff.imass[solubles] = mixed.imass[solubles] * split
        sludge.imass[solubles] = mixed.imass[solubles] - eff.imass[solubles]
        mc = sludge.imass['Water'] / sludge.F_mass
        return mc-target_mc


    def _run(self):
        eff, sludge = self.outs
        solubles, solids = self.solubles, self.solids

        mixed = self._mixed
        mixed.mix_from(self.ins)
        eff.T = sludge.T = mixed.T

        if mixed.F_mass == 0:
            eff.empty()
            sludge.empty()
            self.SKIPPED = True
            return

        mc = mixed.imass['Water']/mixed.F_mass
        target_mc = self.sludge_moisture
        if mc < target_mc:
            if self.SKIPPED == False:
                warn(f'The moisture of influent ({mc:.2%}) is smaller than the desired {target_mc:.0%}, '
                     'simulation skipped.')
                self.SKIPPED = True
            sludge.copy_like(mixed)
            eff.empty()
            return

        sludge.copy_flow(mixed, solids, remove=True) # all solids go to sludge
        eff.copy_flow(mixed, solubles)

        flx.IQ_interpolation(
                f=self._mc_at_split, x0=1e-3, x1=1.-1e-3, ytol=1e-3, maxiter=100,
                args=(solubles, mixed, eff, sludge, target_mc),
                checkbounds=False)
        self.SKIPPED = False


    def _cost(self):
        if self.SKIPPED == False:
            pumps = (self.effluent_pump, self.sludge_pump)
            for i in range(2): pumps[i].simulate()
        else:
            self.baseline_purchase_costs.clear()
            self.power_utility.rate = 0


class BeltThickener(SludgeHandling):
    """
    Gravity belt thickener (GBT) designed based on the manufacture specification
    data sheet. [9]_

    Key parameters include:
    
    * Capacity: 80-100 m3/h.
    * Influent solids concentration: 0.2-1%.
    * Sludge cake moisture content: 90-96%.
    * Motor power: 3 (driving motor) and 1.1 (agitator motor) kW.
    * Belt width: 2.5 m.
    * Weight: 2350 kg.
    * Quote price: $3680 ea for three or more sets.

    The bare module (installation) factor is from Table 25 in Humbird et al. [2]_
    (solids handling equipment).

    Parameters
    ----------
    ins : 
        Influent
    outs : 
        * [0] water-rich supernatant
        * [1] solid-rich sludge
    sludge_moisture : float
        Moisture content of the thickened sludge, [wt% water].
    solubles : tuple
        IDs of the soluble chemicals.
        Note that all chemicals that are not included in this tuple and not
        locked as gas phase (i.e., `chemical.locked_state!='g'`) will be
        treated as solids in simulation.
    max_capacity : float
        Maximum hydraulic loading per belt thickener, [m3/h].
    power_demand : float
        Total power demand of each belt thickener, [kW].

    """

    def _init(self, sludge_moisture=0.96, solubles=(),
                 max_capacity=100, power_demand=4.1):
        SludgeHandling._init(self, sludge_moisture=sludge_moisture,
                             solubles=solubles)
        self.max_capacity = max_capacity
        self.power_demand = power_demand


    def _design(self):
        self._N_thickener = N = math.ceil(self._mixed.F_vol/self.max_capacity)
        self.design_results['Number of thickeners'] = N
        self.F_BM['Thickeners'] = 1.7 # ref [2]
        self.baseline_purchase_costs['Thickeners'] = 4000 * N


    def _cost(self):
        super()._cost()
        self.power_utility.rate += self.power_demand * self.N_thickener


    @property
    def N_thickener(self):
        """[int] Number of required belt thickeners."""
        return self._N_thickener


class SludgeCentrifuge(SludgeHandling, bst.SolidsCentrifuge):
    """
    Solid centrifuge for sludge dewatering.

    The simulation (:meth:`~biosteam.Unit._run` method) and costing (:meth:`~biosteam.Unit._cost` method) are based on 
    :class:`~biosteam.SludgeHandling` and the sizing (:meth:`~biosteam.Unit._design` method) is based on 
    :class:`~biosteam.units.solids_separation.SolidsCentrifuge`.

    Parameters
    ----------
    ins : 
        Influent
    outs : 
        * [0] water-rich supernatant
        * [1] solid-rich sludge
    sludge_moisture : float
        Moisture content of the thickened sludge, [wt% water].
    solubles : tuple
        IDs of the soluble chemicals.
        Note that all chemicals that are not included in this tuple and not
        locked as gas phase (i.e., `chemical.locked_state!='g'`) will be
        treated as solids in simulation.
    
    """

    def _init(self, sludge_moisture=0.8, solubles=(),
              centrifuge_type='scroll_solid_bowl'):
        SludgeHandling._init(self, sludge_moisture=sludge_moisture,
                             solubles=solubles)
        self.centrifuge_type = centrifuge_type

    _run = SludgeHandling._run

    _design = bst.SolidsCentrifuge._design

    _cost = SludgeHandling._cost


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
#                          Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from __future__ import annotations
import pandas as pd
import numpy as np
import flexsolve as flx
from copy import copy as copy_
from numba import njit
from math import ceil
from warnings import warn
import biosteam as bst
from typing import Optional, Sequence, Collection, TYPE_CHECKING
from ._unit import Unit
from numpy.typing import NDArray
if TYPE_CHECKING: from ._system import System

__all__ = ('TEA',)


cashflow_columns = ('Depreciable capital [MM$]',
                    'Fixed capital investment [MM$]',
                    'Working capital [MM$]',
                    'Depreciation [MM$]',
                    'Loan [MM$]',
                    'Loan interest payment [MM$]',
                    'Loan payment [MM$]',
                    'Loan principal [MM$]',
                    'Annual operating cost (excluding depreciation) [MM$]',
                    'Sales [MM$]',
                    'Tax [MM$]',
                    'Incentives [MM$]',
                    'Taxed earnings [MM$]',
                    'Forwarded losses [MM$]',
                    'Net earnings [MM$]',
                    'Cash flow [MM$]',
                    'Discount factor',
                    'Net present value (NPV) [MM$]',
                    'Cumulative NPV [MM$]')

# %% Depreciation utilities

@njit(cache=True)
def generate_DDB_schedule(years):
    val = 1.
    arr = np.ones(years)
    factor = 2. / years
    for i in range(years):
        depreciation = val * factor
        arr[i] = depreciation
        val -= depreciation
    return arr
    
@njit(cache=True)
def generate_SYD_schedule(years):
    digit_sum = years * (years + 1.) * 0.5
    arr = np.ones(years)
    for i in range(years):
        arr[i] = (years - i) / digit_sum
    return arr

# %% Utilities for TEA calculations

@njit(cache=True)
def NPV_at_IRR(IRR, cashflow_array, duration_array):
    """Return NPV at given IRR and cashflow data."""
    return (cashflow_array/(1.+IRR)**duration_array).sum()

@njit(cache=True)
def loan_principal_with_interest(loan, interest):
    principal = 0
    k = 1. + interest
    for i in loan:
        principal *= k
        principal += i
    return principal

@njit(cache=True)
def solve_payment(loan_principal, interest, years):
    f = 1 + interest
    fn = f ** years
    return loan_principal * interest * fn / (fn - 1)

@njit(cache=True)
def taxable_earnings_with_fowarded_losses(taxable_cashflow): # Forwards losses to later years to reduce future taxes
    taxed_earnings = taxable_cashflow.copy()
    for i in range(taxed_earnings.size - 1):
        x = taxed_earnings[i]
        if x < 0:
            taxed_earnings[i] = 0
            taxed_earnings[i + 1] += x
    if taxed_earnings[-1] < 0: taxed_earnings[-1] = 0
    return taxed_earnings

@njit(cache=True)
def add_replacement_cost_to_cashflow_array(equipment_installed_cost, 
                                           equipment_lifetime, 
                                           cashflow_array, 
                                           venture_years,
                                           start):
    N_purchases = ceil(venture_years / equipment_lifetime) 
    equipment_lifetime_index = start + equipment_lifetime
    for i in range(1, N_purchases):
        equipment_lifetime_index += equipment_lifetime
        cashflow_array[int(equipment_lifetime_index)] += equipment_installed_cost

def add_all_replacement_costs_to_cashflow_array(unit_capital_cost, cashflow_array, 
                                                venture_years, start,
                                                lang_factor):
    equipment_lifetime = unit_capital_cost.equipment_lifetime
    if equipment_lifetime:
        if lang_factor:
            installed_costs =  {i: j*lang_factor for i, j in unit_capital_cost.purchase_costs.items()}
        else:
            installed_costs = unit_capital_cost.installed_costs
        if isinstance(equipment_lifetime, int):
             add_replacement_cost_to_cashflow_array(sum(installed_costs.values()), 
                                                    equipment_lifetime,
                                                    cashflow_array,
                                                    venture_years,
                                                    start)
        elif isinstance(equipment_lifetime, dict):
            for name, installed_cost in installed_costs.items():
                lifetime = equipment_lifetime.get(name)
                if lifetime:
                    add_replacement_cost_to_cashflow_array(installed_cost, 
                                                           lifetime,
                                                           cashflow_array,
                                                           venture_years,
                                                           start)

@njit(cache=True)
def fill_taxable_and_nontaxable_cashflows_without_loans(
        D, C, S, C_FC, C_WC, FCI, WC, TDC, VOC, FOC, sales,
        startup_time,
        startup_VOCfrac,
        startup_FOCfrac,
        startup_salesfrac,
        construction_schedule,
        start):
    # Cash flow data and parameters
    # C_FC: Fixed capital
    # C_WC: Working capital
    # D: Depreciation
    # C: Annual operating cost (excluding depreciation)
    # S: Sales
    w0 = startup_time % 1
    end_start = start + int(startup_time)
    w1 = 1. - w0
    C[end_start] = (w0 * startup_VOCfrac * VOC + w1 * VOC
                    + w0 * startup_FOCfrac * FOC + w1 * FOC)
    C[start:end_start] = (startup_VOCfrac * VOC 
                          + startup_FOCfrac * FOC)
    S[end_start] = w0 * startup_salesfrac * sales + w1 * sales
    S[start:end_start] = startup_salesfrac * sales
    start1 = end_start + 1
    C[start1:] = VOC + FOC
    S[start1:] = sales
    C_FC[:start] = FCI * construction_schedule
    C_WC[start-1] = WC
    C_WC[-1] = -WC

def taxable_and_nontaxable_cashflows(
        unit_capital_costs,
        D, C, S, C_FC, C_WC, Loan, LP,
        FCI, WC, TDC, VOC, FOC, sales,
        startup_time,
        startup_VOCfrac,
        startup_FOCfrac,
        startup_salesfrac,
        construction_schedule,
        finance_interest,
        finance_years,
        finance_fraction,
        start, years,
        lang_factor,
        accumulate_interest_during_construction,
    ):
    # Cash flow data and parameters
    # C_FC: Fixed capital
    # C_WC: Working capital
    # Loan: Money gained from loan
    # LP: Loan payment
    # D: Depreciation
    # C: Annual operating cost (excluding depreciation)
    # S: Sales
    fill_taxable_and_nontaxable_cashflows_without_loans(
        D, C, S, C_FC, C_WC, FCI, WC, TDC, VOC, FOC, sales,
        startup_time,
        startup_VOCfrac,
        startup_FOCfrac,
        startup_salesfrac,
        construction_schedule,
        start,
    )
    for i in unit_capital_costs:
        add_all_replacement_costs_to_cashflow_array(i, C_FC, years, start, lang_factor)
    if finance_interest:
        interest = finance_interest
        years = finance_years
        Loan[:start] = loan = finance_fraction*(C_FC[:start])
        if accumulate_interest_during_construction:
            loan_principal = loan_principal_with_interest(loan, interest)
        else:
            loan_principal = loan.sum()
        LP[start:start + years] = solve_payment(loan_principal, interest, years)
        taxable_cashflow = S - C - D - LP
        nontaxable_cashflow = D + Loan - C_FC - C_WC 
        if not accumulate_interest_during_construction: 
            nontaxable_cashflow[:start] -= loan * interest 
    else:
        taxable_cashflow = S - C - D
        nontaxable_cashflow = D - C_FC - C_WC
    return taxable_cashflow, nontaxable_cashflow

def NPV_with_sales(
        sales, 
        taxable_cashflow, 
        nontaxable_cashflow,
        depreciation,
        sales_coefficients,
        discount_factors,
        fill_tax_and_incentives,
    ):
    """Return NPV with an additional annualized sales."""
    taxable_cashflow = taxable_cashflow + sales * sales_coefficients
    tax = np.zeros_like(taxable_cashflow, dtype=float)
    incentives = tax.copy()
    fill_tax_and_incentives(
        incentives, 
        taxable_earnings_with_fowarded_losses(taxable_cashflow),
        nontaxable_cashflow, tax, depreciation
    )
    cashflow = nontaxable_cashflow + taxable_cashflow + incentives - tax
    return (cashflow/discount_factors).sum()

# %% Techno-Economic Analysis

_duration_array_cache = {}

class TEA:
    """
    Abstract TEA class for cash flow analysis.
    
    **Abstract methods**
    
    _DPI(installed_equipment_cost) -> float
        Should return the direct permanent investment (DPI) given the
        installed equipment cost.
    _TDC(DPI) -> float
        Should take direct permanent investment (DPI) as an argument
        and return total depreciable capital (TDC).
    _FCI(TDC) -> float
        Should take total depreciable capital (TDC) as an argument and return
        fixed capital investment (FCI).
    _FOC(FCI) -> float
        Should take fixed capital investment (FCI) as an arguments and return
        fixed operating cost without depreciation (FOC). 
    _fill_tax_and_incentives(incentives, taxable_cashflow, nontaxable_cashflow, tax)
        Should take two empty 1d arrays and fill them with incentive and tax cash flows.
        Additional parameters include taxable_cashflow (sales - costs - 
        depreciation - payments), and nontaxable_cashflow (depreciation - capital 
        cost - working capital).
    
    Parameters
    ----------
    system : 
        Should contain feed and product streams.
    IRR : 
        Internal rate of return (fraction).
    duration : 
        Start and end year of venture (e.g. (2018, 2038)).
    depreciation : 
        Depreciation schedule array or a string with format '{schedule}{years}', 
        where years is the number of years until the property value is zero 
        and schedule is one of the following: 'MACRS' (Modified Accelerated Cost Recovery System), 
        'SL' (straight line), 'DDB' (double declining balance), or 
        'SYD' (sum-of-the-years' digits). If years is not given, it defaults 
        to the venture years at run time. 
    operating_days :  
        Number of operating days per year.
    income_tax : 
        Combined federal and state income tax rate (fraction).
    lang_factor : 
        Lang factor for getting fixed capital investment
        from total purchase cost. If no lang factor, estimate
        capital investment using bare module factors.
    construction_schedule : 
        Construction investment fractions per year (e.g. (0.5, 0.5) for 50%
        capital investment in the first year and 50% investment in the second).
    startup_months : 
        Startup time in months.
    startup_FOCfrac : 
        Fraction of fixed operating costs incurred during startup.
    startup_VOCfrac : 
        Fraction of variable operating costs incurred during startup.
    startup_salesfrac : 
        Fraction of sales achieved during startup.
    WC_over_FCI : 
        Working capital as a fraction of fixed capital investment.
    finance_interest : 
        Yearly interest of capital cost financing as a fraction.
    finance_years : 
        Number of years the loan is paid for.
    finance_fraction :
        Fraction of capital cost that needs to be financed.
    
    Warning
    -------
    When using a Lang factor, a short-cut to get the FCI, we cannot work 
    backwards to get installed equipment cost. For practical purposes, the
    code assumes that installed equipment cost, total depreciable capital 
    (TDC), and fixed capital investment (FCI) are the same when the Lang
    factor is in use. In actuality, the installed equipment cost should be 
    less than the fixed capital investment. 
    
    Examples
    --------
    :doc:`../tutorial/Techno-economic_analysis` 

    """
    __slots__ = ('system', 'income_tax', 'WC_over_FCI',
                 'finance_interest', 'finance_years', 'finance_fraction',
                 '_construction_schedule', '_startup_time',
                 'startup_FOCfrac', 'startup_VOCfrac', 'startup_salesfrac',
                 '_startup_schedule', '_operating_days',
                 '_duration', '_depreciation_key', '_depreciation',
                 '_years', '_duration', '_start',  'IRR', '_IRR', '_sales',
                 '_duration_array_cache', 'accumulate_interest_during_construction')
    
    #: Available depreciation schedules. Defaults include modified 
    #: accelerated cost recovery system from U.S. IRS publication 946 (MACRS),
    #: half-year convention.
    depreciation_schedules: dict[tuple(str, int), NDArray[float]] = {
        ('MACRS', 3): np.array([.3333, .4445, .1481, .0741]),

        ('MACRS', 5): np.array([.2000, .3200, .1920,
                                .1152, .1152, .0576]),
              
        ('MACRS', 7):  np.array([.1429, .2449, .1749,
                                 .1249, .0893, .0892,
                                 .0893, .0446]),
        
        ('MACRS', 10): np.array([.1000, .1800, .1440,
                                 .1152, .0922, .0737,
                                 .0655, .0655, .0656,
                                 .0655, .0328]),
      
        ('MACRS', 15): np.array([.0500, .0950, .0855,
                                 .0770, .0693, .0623,
                                 .0590, .0590, .0591,
                                 .0590, .0591, .0590,
                                 .0591, .0590, .0591,
                                 .0295]),
      
        ('MACRS', 20): np.array([0.03750, 0.07219, 0.06677,
                                 0.06177, 0.05713, 0.05285,
                                 0.04888, 0.04522, 0.04462,
                                 0.04461, 0.04462, 0.04461,
                                 0.04462, 0.04461, 0.04462,
                                 0.04461, 0.04462, 0.04461,
                                 0.04462, 0.04461, 0.02231])
    }
    
    #: Investment site factors used to multiply the total permanent 
    #: investment (TPI), also known as total fixed capital (FCI), to 
    #: account for locality cost differences based on labor availability,
    #: workforce efficiency, local rules, etc.
    investment_site_factors: dict[str, float] = {
        'U.S. Gulf Coast': 1.0,
        'U.S. Southwest': 0.95,
        'U.S. Northwest': 1.10,
        'U.S. Midwest': 1.15,
        'U.S. West Coast': 1.25,
        'Western Europe': 1.2,
        'Mexico': 0.95,
        'Japan': 1.15,
        'Pacific Rim': 1.0,
        'India': 0.85,
    }

    class Accounting:
        __slots__ = ('index', 'data', 'names', 'units')
        
        def __init__(self, units, names=None):
            self.units = units
            self.names = names
            self.index = []
            self.data =[]
        
        def entry(self, index: str, cost: list|float, notes: str = ''):
            self.index.append(index)
            if getattr(cost, 'ndim', 0) == 0: cost = float(cost)
            if isinstance(cost, (float, int, str)):
                self.data.append([notes, cost])
            else:
                self.data.append([notes, *cost])

        @property
        def total_costs(self):
            if self.names is None:
                return sum([i[1] for i in self.data])
            else:
                N = len(self.data[0])
                return np.array([sum([i[index] for i in self.data]) for index in range(1, N)])
        
        def table(self):
            names = self.names
            units = self.units
            index = self.index
            return pd.DataFrame(
                self.data, 
                index=pd.MultiIndex.from_tuples(index) if isinstance(index[0], tuple) else index,
                columns=('Notes', f'Cost [{units}]') if names is None else ('Notes', *[f'{i}\n[{units}]' for i in names]),
            )

    def __init_subclass__(cls, isabstract=False):
        if isabstract: return
        for method in ('_DPI', '_TDC', '_FCI', '_FOC'):
            if not hasattr(cls, method):
                breakpoint()
                raise NotImplementedError(
                    f"subclass must implement a '{method}' method unless the "
                     "'isabstract' keyword argument is True"
                )

    def copy(self, system=None):
        """Create a copy."""
        new = copy_(self)
        if system is not None:
            new.system = system
            system._TEA = new
        return new

    def __init__(self, system: System, IRR: float, duration: tuple[int, int], 
                 depreciation: str|NDArray[float], income_tax: float,
                operating_days: float, lang_factor: float|None, 
                construction_schedule: Sequence[float],
                startup_months: float, startup_FOCfrac: float, startup_VOCfrac: float,
                startup_salesfrac: float, WC_over_FCI: float,  finance_interest: float,
                finance_years: int, finance_fraction: float,
                accumulate_interest_during_construction: bool=False):
        #: System being evaluated.
        self.system: System = system
        
        self.duration = duration
        self.depreciation = depreciation
        self.construction_schedule = construction_schedule
        self.startup_months = startup_months
        self.operating_days = operating_days
        
        #: Internal rate of return (fraction).
        self.IRR: float = IRR
        
        #: Combined federal and state income tax rate (fraction).
        self.income_tax: float = income_tax
        
        self.lang_factor = lang_factor
        
        #: Fraction of fixed operating costs incurred during startup.
        self.startup_FOCfrac: float = startup_FOCfrac
        
        #: Fraction of variable operating costs incurred during startup.
        self.startup_VOCfrac: float = startup_VOCfrac
        
        #: Fraction of sales achieved during startup.
        self.startup_salesfrac: float = startup_salesfrac
        
        #: Working capital as a fraction of fixed capital investment.
        self.WC_over_FCI: float = WC_over_FCI
        
        #: Yearly interest of capital cost financing as a fraction.
        self.finance_interest: float = finance_interest
        
        #: Number of years the loan is paid for.
        self.finance_years: int = finance_years
        
        #: Fraction of capital cost that needs to be financed.
        self.finance_fraction: float = finance_fraction
        
        #: Guess IRR for solve_IRR method
        self._IRR: float = IRR
        
        #: Guess cost for solve_price method
        self._sales: float = 0
        
        #: Whether to immediately pay interest before operation or to accumulate interest during construction
        self.accumulate_interest_during_construction = accumulate_interest_during_construction
        
        #: For convenience, set a TEA attribute for the system
        system._TEA = self

    def _get_duration(self):
        return (self._start, self._years)

    def _DPI(self, installed_equipment_cost):
        return installed_equipment_cost # For compatibility with Lang factors

    def _TDC(self, DPI):
        return DPI # For compatibility with Lang factors
    
    def _FCI(self, TDC):
        return TDC # For compatibility with Lang factors

    @property
    def save_report(self):
        return self.system.save_report

    @property
    def units(self) -> set[Unit]:
        """All unit operations with costs."""
        return self.system.cost_units  

    @property
    def feeds(self) -> list[Unit]:
        """All feed streams."""
        return self.system.feeds  
      
    @property
    def products(self) -> list[Unit]:
        """All product streams."""
        return self.system.products

    @property
    def operating_days(self) -> float:
        """Number of operating days per year."""
        return self.system.operating_hours / 24
    @operating_days.setter
    def operating_days(self, days):
        """Number of operating days per year."""
        self.operating_hours = 24 * days
    
    @property
    def operating_hours(self) -> float:
        """Number of operating hours per year."""
        return self.system.operating_hours
    @operating_hours.setter
    def operating_hours(self, hours):
        self.system.operating_hours = hours
    
    @property
    def lang_factor(self) -> float|None:
        """
        Lang factor for getting fixed capital investment from 
        total purchase cost. If no lang factor, estimate capital investment 
        using bare module factors.
        
        """
        return self.system.lang_factor
    @lang_factor.setter
    def lang_factor(self, lang_factor):
        self.system.lang_factor = lang_factor
    
    @property
    def duration(self) -> tuple[int, int]:
        """Start and end year of venture."""
        return self._duration
    @duration.setter
    def duration(self, duration):
        start, end = [int(i) for i in duration]
        self._duration = (start, end)
        self._years = end - start

    @property
    def depreciation(self) -> str|NDArray[float]:
        """
        Depreciation schedule array or a string with format '{schedule}{years}', 
        where years is the number of years until the property value is zero 
        and schedule is one of the following: 'MACRS' (Modified Accelerated Cost Recovery System), 
        'SL' (straight line), 'DDB' (double declining balance), or 
        'SYD' (sum-of-the-years' digits). If years is not given, it defaults 
        to the venture years at run time.
        """
        return self._depreciation
    @depreciation.setter
    def depreciation(self, depreciation):
        if isinstance(depreciation, str):
            self._depreciation_key = self._depreciation_key_from_name(depreciation)
            self._depreciation = depreciation
        else:
            try:
                self._depreciation = np.array(depreciation, dtype=float)
            except:
                raise TypeError(
                   f"invalid depreciation type '{type(depreciation).__name__}'; "
                    "depreciation must be either an array or a string"
                ) from None
            else:
                self._depreciation_key = None
    
    @classmethod
    def _depreciation_key_from_name(cls, name):
        for prefix in ('MACRS', 'SL', 'DDB', 'SYD'):
            if name.startswith(prefix):
                years = name[len(prefix):]
                key = (prefix, int(years) if years else None)
                if prefix == 'MACRS' and key not in cls.depreciation_schedules:
                    raise ValueError(
                        f"depreciation name {repr(name)} has a valid "
                         "format, but is not yet implemented in BioSTEAM"
                    )
                return key
        raise ValueError(
               f"invalid depreciation name {repr(name)}; "
                "name must have format '{schedule}{years}', "
                "where years is the number of years until the property value is zero "
                "and schedule is one of the following: 'MACRS' (Modified Accelerated Cost Recovery System), "
                "'SL' (straight line), 'DDB' (double declining balance), or "
                "'SYD' (sum-of-the-years' digits)"
            )

    @classmethod
    def _depreciation_array_from_key(cls, key):
        depreciation_schedules = cls.depreciation_schedules
        if key in depreciation_schedules:
            return depreciation_schedules[key]
        else:
            schedule, years = key
            if schedule == 'SL':
                arr = np.full(years, 1./years)
            elif schedule == 'DDB':
                arr = generate_DDB_schedule(years)
            elif schedule == 'SYD':
                arr = generate_SYD_schedule(years)
            else: # pragma: no cover
                raise RuntimeError(f'unknown depreciation schedule {repr(schedule)}')
            depreciation_schedules[key] = arr
            return arr
    
    @property
    def construction_schedule(self) -> Sequence[float]:
        """Construction investment fractions per year, starting from year 0.
        For example, for 50% capital investment in year 0 and 50% investment 
        in year 1, use (0.5, 0.5)."""
        return self._construction_schedule
    @construction_schedule.setter
    def construction_schedule(self, schedule):
        self._construction_schedule = np.array(schedule, dtype=float)
        self._start = len(schedule)
    
    @property
    def startup_months(self) -> float:
        return self._startup_time * 12.
    @startup_months.setter
    def startup_months(self, months):
        self._startup_time = months / 12.
    
    @property
    def sales(self) -> float:
        """Total sales [USD/yr]."""
        return self.system.sales
    @property
    def material_cost(self) -> float:
        """Total material cost [USD/yr]."""
        return self.system.material_cost
    @property
    def utility_cost(self) -> float:
        """Total utility cost [USD/yr]."""
        return self.system.utility_cost
    @property
    def purchase_cost(self):
        """Total purchase cost [USD]."""
        return self.system.purchase_cost
    @property
    def installed_equipment_cost(self) -> float:
        """Total installed cost [USD]."""
        return self.system.installed_equipment_cost
    @property
    def DPI(self) -> float:
        """Direct permanent investment [USD]."""
        return self._DPI(self.installed_equipment_cost)
    @property
    def TDC(self) -> float:
        """Total depreciable capital [USD]."""
        return self._TDC(self.DPI)
    @property
    def FCI(self) -> float:
        """Fixed capital investment [USD]."""
        return self._FCI(self.TDC)
    @property
    def TCI(self) -> float:
        """Total capital investment [USD]."""
        return (1. + self.WC_over_FCI)*self.FCI
    @property
    def FOC(self) -> float:
        """Fixed operating costs [USD/yr]."""
        return self._FOC(self.FCI)
    @property
    def VOC(self) -> float:
        """Variable operating costs [USD/yr]."""
        return self.material_cost + self.utility_cost
    @property
    def AOC(self) -> float:
        """Annual operating cost excluding depreciation [USD/yr]."""
        return self.FOC + self.VOC
    @property
    def working_capital(self) -> float:
        return self.WC_over_FCI * self.FCI
    
    @property
    def annual_depreciation(self) -> float:
        """Depreciation [USD/yr] equivalent to TDC divided by the duration of the venture."""
        return self.TDC/(self.duration[1]-self.duration[0])

    @property
    def ROI(self) -> float:
        """Return on investment [1/yr] without accounting for annualized depreciation."""
        return self.net_earnings / self.TCI
    @property
    def net_earnings(self) -> float:
        """Net earnings without accounting for annualized depreciation."""
        net_earnings = self.sales - self.AOC
        if net_earnings < 0:
            return net_earnings
        else:
            return (1 - self.income_tax) * net_earnings
    @property
    def PBP(self) -> float:
        """Pay back period [yr] without accounting for annualized depreciation."""
        FCI = self.FCI
        net_earnings = self.net_earnings
        return FCI/net_earnings

    def _get_duration_array(self):
        key = start, years = (self._start, self._years)
        if key in _duration_array_cache:
            duration_array = _duration_array_cache[key]
        else:
            if len(_duration_array_cache) > 100: _duration_array_cache.clear()
            _duration_array_cache[key] = duration_array = np.arange(-start+1, years+1, dtype=float)
        return duration_array

    def _get_depreciation_array(self):
        key = self._depreciation_key
        if key is None: 
            return self._depreciation
        else:
            schedule, years = self._depreciation_key
            if years is None:
                years = self._years
                key = (schedule, years)
            return self._depreciation_array_from_key(key)
            
    def _fill_depreciation_array(self, D, start, years, TDC):
        depreciation_array = self._get_depreciation_array()
        N_depreciation_years = depreciation_array.size
        if N_depreciation_years > years:
            raise RuntimeError('depreciation schedule is longer than plant lifetime')
        D[start:start + N_depreciation_years] = TDC * depreciation_array

    def get_cashflow_table(self):
        """Return DataFrame of the cash flow analysis."""
        # Cash flow data and parameters
        # index: Year since construction until end of venture
        # C_D: Depreciable capital
        # C_FC: Fixed capital
        # C_WC: Working capital
        # D: Depreciation
        # L: Loan revenue
        # LI: Loan interest payment
        # LP: Loan payment
        # LPl: Loan principal
        # C: Annual operating cost (excluding depreciation)
        # S: Sales
        # T: Tax
        # I: Incentives
        # TE: Taxed earnings
        # FL: Forwarded losses
        # NE: Net earnings
        # CF: Cash flow
        # DF: Discount factor
        # NPV: Net present value
        # CNPV: Cumulative NPV
        TDC = self.TDC
        FCI = self._FCI(TDC)
        start = self._start
        years = self._years
        FOC = self._FOC(FCI)
        VOC = self.VOC
        sales = self.sales
        length = start + years
        C_D, C_FC, C_WC, D, L, LI, LP, LPl, C, S, T, I, TE, FL, NE, CF, DF, NPV, CNPV = data = np.zeros((19, length))
        self._fill_depreciation_array(D, start, years, TDC)
        w0 = self._startup_time % 1
        w1 = 1. - w0
        end_start = start + int(self._startup_time)
        C[end_start] = (w0 * self.startup_VOCfrac * VOC + w1 * VOC
                        + w0 * self.startup_FOCfrac * FOC + w1 * FOC)
        C[start:end_start] = (self.startup_VOCfrac * VOC 
                              + self.startup_FOCfrac * FOC)
        S[end_start] = w0 * self.startup_salesfrac * sales + w1 * sales
        S[start:end_start] = self.startup_salesfrac * sales
        start1 = end_start + 1
        C[start1:] = VOC + FOC
        S[start1:] = sales
        WC = self.WC_over_FCI * FCI
        C_D[:start] = TDC*self._construction_schedule
        C_FC[:start] = FCI*self._construction_schedule
        C_WC[start-1] = WC
        C_WC[-1] = -WC
        system = self.system
        lang_factor = system.lang_factor
        unit_capital_costs = system.unit_capital_costs.values() if isinstance(system, bst.AgileSystem) else system.cost_units
        for i in unit_capital_costs: add_all_replacement_costs_to_cashflow_array(i, C_FC, years, start, lang_factor)
        if self.finance_interest:
            interest = self.finance_interest
            years = self.finance_years
            end = start + years
            L[:start] = loan = self.finance_fraction*(C_FC[:start])
            accumulate_interest_during_construction = self.accumulate_interest_during_construction
            if accumulate_interest_during_construction:
                initial_loan_principal = loan_principal_with_interest(loan, interest)
            else:
                initial_loan_principal = loan.sum()
            LP[start:end] = solve_payment(initial_loan_principal, interest, years)
            loan_principal = 0
            if accumulate_interest_during_construction:
                for i in range(end):
                    LI[i] = li = (loan_principal + L[i]) * interest 
                    LPl[i] = loan_principal = loan_principal - LP[i] + li + L[i]
            else:
                for i in range(end):
                    if i < start: 
                        li = 0
                    else:
                        li = (loan_principal + L[i]) * interest 
                    LI[i] = li
                    LPl[i] = loan_principal = loan_principal - LP[i] + li + L[i]
                LI[:start] = L[:start] * interest # Interest still needs to be payed
            taxable_cashflow = S - C - D - LP
            nontaxable_cashflow = D + L - C_FC - C_WC
            if not accumulate_interest_during_construction:
                nontaxable_cashflow[:start] -= LI[:start] # Subtract the interest during construction from NPV
        else:
            taxable_cashflow = S - C - D
            nontaxable_cashflow = D - C_FC - C_WC
        TE[:] = taxable_earnings_with_fowarded_losses(taxable_cashflow)
        FL[1:] = (taxable_cashflow - TE).cumsum()[:-1]
        self._fill_tax_and_incentives(
            I, TE, nontaxable_cashflow, T, D
        )
        NE[:] = taxable_cashflow + I - T
        CF[:] = NE + nontaxable_cashflow
        DF[:] = 1/(1.+self.IRR)**self._get_duration_array()
        NPV[:] = CF * DF
        CNPV[:] = NPV.cumsum()
        DF *= 1e6
        data /= 1e6
        return pd.DataFrame(data.transpose(),
                            index=np.arange(self._duration[0]-start, self._duration[1]),
                            columns=cashflow_columns)
    @property
    def NPV(self) -> float:
        """Net present value."""
        taxable_cashflow, nontaxable_cashflow, depreciation = self._taxable_nontaxable_depreciation_cashflows()
        tax = np.zeros_like(taxable_cashflow)
        incentives = tax.copy()
        self._fill_tax_and_incentives(
            incentives,
            taxable_earnings_with_fowarded_losses(taxable_cashflow),
            nontaxable_cashflow, tax, depreciation
        )
        cashflow = nontaxable_cashflow + taxable_cashflow + incentives - tax
        return NPV_at_IRR(self.IRR, cashflow, self._get_duration_array())
    
    def _AOC(self, FCI):
        """Return AOC at given FCI"""
        return self._FOC(FCI) + self.VOC
    
    def _taxable_nontaxable_depreciation_cashflows(self):
        """Return taxable, nontaxable and depreciation cash flows by year as a tuple[1d array, 1d array, 1d array]."""
        # Cash flow data and parameters
        # C_FC: Fixed capital
        # C_WC: Working capital
        # Loan: Money gained from loan
        # LP: Loan payment
        # D: Depreciation
        # C: Annual operating cost (excluding depreciation)
        # S: Sales
        # NE: Net earnings
        # CF: Cash flow
        TDC = self.TDC
        FCI = self._FCI(TDC)
        start = self._start
        years = self._years
        FOC = self._FOC(FCI)
        VOC = self.VOC
        D, C_FC, C_WC, Loan, LP, C, S = np.zeros((7, start + years))
        self._fill_depreciation_array(D, start, years, TDC)
        WC = self.WC_over_FCI * FCI
        system = self.system
        return (
            *taxable_and_nontaxable_cashflows(
                system.unit_capital_costs if isinstance(system, bst.AgileSystem) else system.cost_units,
                D, C, S, C_FC, C_WC, Loan, LP,
                FCI, WC, TDC, VOC, FOC, self.sales,
                self._startup_time,
                self.startup_VOCfrac,
                self.startup_FOCfrac,
                self.startup_salesfrac,
                self._construction_schedule,
                self.finance_interest,
                self.finance_years,
                self.finance_fraction,
                start, years,
                self.lang_factor,
                self.accumulate_interest_during_construction,
            ),
            D
        )
    
    def _fill_tax_and_incentives(self, incentives, taxable_cashflow, nontaxable_cashflow, tax, depreciation):
        tax[:] = self.income_tax * taxable_cashflow
    
    def _net_earnings_and_nontaxable_cashflow_arrays(self):
        taxable_cashflow, nontaxable_cashflow, depreciation = self._taxable_nontaxable_depreciation_cashflows()
        size = taxable_cashflow.size
        tax = np.zeros(size)
        incentives = tax.copy()
        self._fill_tax_and_incentives(
            incentives, 
            taxable_earnings_with_fowarded_losses(taxable_cashflow), 
            nontaxable_cashflow, tax, depreciation
        )
        net_earnings = taxable_cashflow + incentives - tax
        return net_earnings, nontaxable_cashflow
    
    @property
    def cashflow_array(self) -> NDArray[float]:
        """Cash flows by year."""
        return sum(self._net_earnings_and_nontaxable_cashflow_arrays())
    
    @property
    def net_earnings_array(self) -> NDArray[float]:
        """Net earnings by year."""
        return self._net_earnings_and_nontaxable_cashflow_arrays()[0]
    
    def production_costs(self, products: Sequence[bst.Stream], with_annual_depreciation: Optional[bool]=True):
        """
        Return production cost of products [USD/yr].
        
        Parameters
        ----------
        products : 
            Main products of the system.
        with_annual_depreciation: 
            Whether to add annualized depreciation to the production costs.
        
        Notes
        -----
        If there is more than one main product, The production cost is
        proportionally allocated to each of the main products with respect to
        their marketing values. The marketing value of each product is
        determined by the annual production multiplied by its selling price.
        
        """
        system = self.system
        market_values = np.array([system.get_market_value(i) for i in products])
        total_market_value = market_values.sum()
        weights = market_values/total_market_value
        return weights * self.total_production_cost(products, with_annual_depreciation)
        
    def total_production_cost(self, products: Collection[bst.Stream], with_annual_depreciation: Optional[bool]=True):
        """
        Return total production cost of products [USD/yr].
        
        Parameters
        ----------
        products : 
            Main products of the system.
        with_annual_depreciation: 
            Whether to add annualized depreciation to the production costs.
        
        """
        system = self.system
        coproduct_sales = self.sales - sum([system.get_market_value(i) for i in products])
        if with_annual_depreciation:
            TDC = self.TDC
            annual_depreciation = TDC/(self.duration[1]-self.duration[0])
            AOC = self._AOC(self._FCI(TDC))
            return AOC - coproduct_sales + annual_depreciation
        else:
            return self.AOC - coproduct_sales
    
    def solve_IRR(self, financing=True, bounds=None):
        """Return the IRR at the break even point (NPV = 0) through cash flow analysis."""
        IRR = self._IRR
        if not IRR or np.isnan(IRR) or IRR < 0.: IRR = 0.01
        if financing:
            args = (self.cashflow_array, self._get_duration_array())
            if bounds:
                IRR = flx.IQ_interpolation(
                    NPV_at_IRR, *bounds, x=IRR, xtol=1e-6, ytol=10.,
                    maxiter=200, args=args, checkiter=False
                )
            else:
                IRR = flx.aitken_secant(
                    NPV_at_IRR, IRR, 1.0001 * IRR + 1e-3, xtol=1e-6, ytol=10.,
                    maxiter=200, args=args, checkiter=False
                )
        else:
            financing_values = self.finance_fraction, self.finance_interest
            self.finance_fraction = self.finance_interest = None
            try:
                args = (self.cashflow_array, self._get_duration_array())
                if bounds:
                    IRR = flx.IQ_interpolation(
                        NPV_at_IRR, *bounds, x=IRR, xtol=1e-6, ytol=10.,
                        maxiter=200, args=args, checkiter=False
                    )
                else:
                    IRR = flx.aitken_secant(
                        NPV_at_IRR, IRR, 1.0001 * IRR + 1e-3, xtol=1e-6, ytol=10.,
                        maxiter=200, args=args, checkiter=False
                    )
            finally:
                self.finance_fraction, self.finance_interest = financing_values
        self._IRR = IRR
        return IRR
        
    def solve_price(self, streams: bst.Stream|Collection[bst.Stream]):
        """
        Return the price [USD/kg] of a stream(s) at the break even point (NPV = 0)
        through cash flow analysis. 
        
        Parameters
        ----------
        streams :
            Streams with variable selling price.
            
        """
        if isinstance(streams, bst.Stream): streams = [streams]
        system = self.system
        price2cost = sum([system._price2cost(i) for i in streams])
        if price2cost == 0.: raise ValueError('cannot solve price of empty streams')
        try:
            sales = self.solve_sales()
        except:
            original_prices = [i.price for i in streams]
            for i in streams: i.price = 0.
            sales = self.solve_sales()
            current_price = 0.
            for i, j in zip(streams, original_prices): i.price = j 
        else:
            current_price = sum([system.get_market_value(i) for i in streams]) / abs(price2cost)
        return current_price + sales / price2cost 
        
    def VOC_table(
            self, products, functional_unit='MT', with_products=False, 
        ):
        return bst.report.voc_table(
            [self.system], products, 
            system_names=None, functional_unit=functional_unit,
            with_products=with_products, dataframe=True
        )
        
    def CAPEX_table(self):
        return NotImplemented
    
    def FOC_table(self):
        return NotImplemented
    
    def solve_sales(self):
        """
        Return the required additional sales [USD] to reach the breakeven 
        point (NPV = 0) through cash flow analysis. 
        
        """
        discount_factors = (1 + self.IRR)**self._get_duration_array()
        sales_coefficients = np.ones_like(discount_factors, dtype=float)
        start = self._start
        sales_coefficients[:start] = 0
        w0 = self._startup_time % 1
        end_start = start + int(self._startup_time)
        sales_coefficients[end_start] = w0 * self.startup_salesfrac + (1. - w0) 
        sales_coefficients[start:end_start] = self.startup_salesfrac
        taxable_cashflow, nontaxable_cashflow, depreciation = self._taxable_nontaxable_depreciation_cashflows()
        if np.isnan(taxable_cashflow).any():
            warn('nan encountered in cashflow array; resimulating system', category=RuntimeWarning)
            self.system.simulate()
            taxable_cashflow, nontaxable_cashflow, depreciation = self._taxable_nontaxable_depreciation_cashflows()
            if np.isnan(taxable_cashflow).any():
                raise RuntimeError('nan encountered in cashflow array')
        args = (taxable_cashflow, 
                nontaxable_cashflow, 
                depreciation,
                sales_coefficients,
                discount_factors,
                self._fill_tax_and_incentives)
        x0 = self._sales if np.isfinite(self._sales) else 0
        f = NPV_with_sales
        y0 = f(x0, *args)
        x1 = x0 - y0 / self._years # First estimate
        try:
            sales = flx.aitken_secant(f, x0, x1, xtol=10, ytol=100.,
                                      maxiter=1000, args=args, checkiter=True)
        except:
            bracket = flx.find_bracket(f, x0, x1, args=args)
            sales = flx.IQ_interpolation(f, *bracket, args=args, xtol=10, ytol=100, maxiter=1000, checkiter=False)
        self._sales = sales
        return sales
    
    def __repr__(self):
        return f'{type(self).__name__}({self.system.ID}, ...)'
    
    def _info(self):
        return (f'{type(self).__name__}: {self.system}\n'
                f'NPV: {self.NPV:,.0f} USD at {self.IRR:.1%} IRR')
    
    def show(self):
        """Prints information on unit."""
        print(self._info())
    _ipython_display_ = show


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2024, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autoclass:: biosteam.units.nrel_bioreactor.NRELBatchBioreactor
.. autoclass:: biosteam.units.nrel_bioreactor.NRELFermentation

References
----------
.. [1] D. Humbird, R. Davis, L. Tao, C. Kinchin, D. Hsu, and A. Aden
    National. Renewable Energy Laboratory Golden, Colorado. P. Schoen,
    J. Lukas, B. Olthof, M. Worley, D. Sexton, and D. Dudgeon. Harris Group
    Inc. Seattle, Washington and Atlanta, Georgia. Process Design and Economics
    for Biochemical Conversion of Lignocellulosic Biomass to Ethanol Dilute-Acid
    Pretreatment and Enzymatic Hydrolysis of Corn Stover. May 2011. Technical
    Report NREL/TP-5100-47764

.. [2] Oliveira, Samuel C., et al. "Discrimination between ethanol 
    inhibition models in a continuous alcoholic fermentation process using
    flocculating yeast." Applied biochemistry and biotechnology 74.3 (1998): 161-172.

.. [3] Oliveira, Samuel C., et al. "Continuous ethanol fermentation in a
    tower reactor with flocculating yeast recycle: scale-up effects on process
    performance, kinetic parameters and model predictions." Bioprocess
    Engineering 20.6 (1999): 525-530.

.. [4] Oliveira, Samuel C., et al. "Mathematical modeling of a continuous
    alcoholic fermentation process in a two-stage tower reactor cascade with
    flocculating yeast recycle." Bioprocess and biosystems engineering 38.3
    (2015): 469-479.

.. [5] Oliveira, Samuel C., et al. "Kinetic Modeling of 1‐G Ethanol
    Fermentations." Fermentation Processes. InTech, 2017.
    
    
"""
import numpy as np
from .. import Unit
from .design_tools import size_batch
from .decorators import cost
from math import ceil
from scipy.integrate import odeint
from thermosteam.reaction import Reaction, ParallelReaction

__all__ = (
    'NRELBatchBioreactor', 'NRELFermentation',
    'BatchBioreactor', 'Fermentation', # For backwards compatibility
) 

# %% NREL

@cost('Recirculation flow rate', 'Recirculation pumps', kW=30, S=77.22216,
      cost=47200, n=0.8, BM=2.3, CE=522, N='Number of reactors')
@cost('Reactor volume', 'Cleaning in place', CE=521.9,
      cost=421e3, S=3785, n=0.6, BM=1.8)
@cost('Reactor volume', 'Agitators', CE=521.9, cost=52500,
      S=3785, n=0.5, kW=22.371, BM=1.5, N='Number of reactors')
@cost('Reactor volume', 'Reactors', CE=521.9, cost=844000,
      S=3785, n=0.5, BM=1.5, N='Number of reactors')
@cost('Reactor duty', 'Heat exchangers', CE=522, cost=23900,
      S=20920000.0, n=0.7, BM=2.2, N='Number of reactors',
      magnitude=True) # Based on a similar heat exchanger
class NRELBatchBioreactor(Unit, isabstract=True):
    """
    Abstract Bioreactor class. Conversion is based on reaction time, `tau`.
    Cleaning and unloading time,`tau_0`, fraction of working volume, `V_wf`,
    and number of reactors, `N_reactors`, are attributes that can be changed.
    The cost of a reactor is based on the NREL batch fermentation tank cost 
    assuming volumetric scaling with a 6/10th exponent [1]_. 
    
    **Abstract methods**
    
    _run():
        Must set outlet stream flows.
        
    Parameters
    ----------
    ins : 
        Inlet fluids to be mixed into the reactor.
    outs : 
        * [0] Vent
        * [1] Effluent
    tau : float
        Reaction time [hr].
    N : int, optional
        Number of batch reactors
    V : float, optional
        Target volume of reactors [m^3].
    T=305.15 : float
        Operating temperature of reactor [K].
    P=101325 : float
        Operating pressure of reactor [Pa].
    Nmin=2 : int
        Minimum number of fermentors.
    Nmax=36: int
        Maximum number of fermentors.  
    
    Notes
    -----
    Either N or V must be given.
    
    """
    _units = {'Reactor volume': 'm3',
              'Cycle time': 'hr',
              'Batch time': 'hr',
              'Loading time': 'hr',
              'Total dead time': 'hr',
              'Reactor duty': 'kJ/hr',
              'Recirculation flow rate': 'm3/hr'}
    _N_ins = _N_outs = 2
    
    #: [bool] If True, number of reactors (N) is chosen as to minimize installation cost in every simulation. Otherwise, N remains constant.
    autoselect_N = False
    
    #: [float] Cleaning and unloading time (hr).
    tau_0 = 3
    
    #: [float] Fraction of filled tank to total tank volume.
    V_wf = 0.9
    
    def _get_design_info(self):
        return (('Cleaning and unloading time', self.tau_0, 'hr'),
                ('Working volume fraction', self.V_wf, ''))
    
    def _init(self, tau=None, N=None, V=None, T=305.15, P=101325,
              Nmin=2, Nmax=36):
        self._N = N; self._V = V
        
        #: [float] Reaction time [hr].
        self.tau = tau
        
        #: [int] Number of batch reactors
        if N: self.N = N
        
        #: [float] Target volume of a fermentor
        if V: self.V = V
        
        #: [float] Operating temperature of reactor [K].
        self.T = T
        
        #: [float] Operating pressure of reactor [Pa].
        self.P = P
        
        #: [int] Minimum number of fermentors
        self.Nmin = Nmin
        
        #: [int] Maximum number of fermentors
        self.Nmax = Nmax
        
    def _setup(self):
        super()._setup()
        vent, effluent = self.outs
        vent.phase = 'g'
        vent.T = effluent.T = self.T
        vent.P = effluent.P = self.P
        
    @property
    def vent(self):
        return self.outs[0]
    @property
    def effluent(self):
        return self.outs[1]
        
    @property
    def N(self):
        """[int] Number of reactor."""
        return self._N
    @N.setter
    def N(self, N):
        if N is None: 
            self._N = N
            return
        if N <= 1:
            raise ValueError(f"number of reactors must be greater than 1, value {N} is infeasible")
        assert not self._V, 'cannot specify both reactor volume and number of reactors'
        self._N = ceil(N)

    @property
    def V(self):
        """[float] Reactor volume."""
        return self._V
    @V.setter
    def V(self, V):
        if V <= 1:
            raise ValueError(f"reactor volume must be greater than 1, value {V} is infeasible")
        assert not self._N, 'cannot specify both reactor volume and number of reactors'
        self._V = V

    @property
    def tau(self):
        return self._tau
    @tau.setter
    def tau(self, tau):
        self._tau = tau
    
    @property
    def N_at_minimum_capital_cost(self):
        cost_old = np.inf
        self.autoselect_N = False
        self._N, N = 2, self._N
        cost_new = self.purchase_cost
        self._summary()
        while cost_new < cost_old:
            self._N += 1
            self._summary()
            cost_old = cost_new
            cost_new = self.purchase_cost
        self._N, N = N, self._N
        self.autoselect_N = True
        return N - 1
        
    def _design(self):
        effluent = self.effluent
        v_0 = effluent.F_vol
        tau = self._tau
        tau_0 = self.tau_0
        V_wf = self.V_wf
        Design = self.design_results
        if self.autoselect_N:
            N = self.N_at_minimum_capital_cost
        elif self.V:
            N = v_0 / self.V / V_wf * (tau + tau_0) + 1
            if N < 2:
                N = 2
            else:
                N = ceil(N)
        else:
            N = self._N
        Design.update(size_batch(v_0, tau, tau_0, N, V_wf))
        Design['Number of reactors'] = N
        Design['Recirculation flow rate'] = v_0 / N
        duty = self.Hnet
        Design['Reactor duty'] = duty
        self.add_heat_utility(duty, self.T)


class NRELFermentation(NRELBatchBioreactor):
    """
    Create a Fermentation object which models large-scale batch fermentation
    for the production of 1st generation ethanol using yeast
    [2]_ [3]_ [4]_ [5]_. A compound with CAS 'Yeast' must be present.
    Only sucrose and glucose are taken into account for conversion.
    Conversion is based on reaction time, `tau`. Cleaning and unloading time,
    `tau_0`, fraction of working volume, `V_wf`, and number of reactors,
    `N_reactors`, are attributes that can be changed. Cost of a reactor
    is based on the NREL batch fermentation tank cost assuming volumetric
    scaling with a 6/10th exponent [1]_. 
    
    Parameters
    ----------
    ins : 
        Inlet fluids to be mixed into the fermentor.
    outs : 
        * [0] Vent
        * [1] Effluent
    tau : float
        Reaction time.
    N : int, optional
        Number of batch reactors
    V : float, optional
        Target volume of reactors [m^3].
    T=305.15 : float
        Temperature of reactor [K].
    P=101325 : float
        Operating pressure of reactor [Pa].
    Nmin=2 : int
        Minimum number of fermentors.
    Nmax=36: int
        Maximum number of fermentors.  
    efficiency=0.9 : float, optional
        User enforced efficiency.
    iskinetic=False: bool, optional
        If True, `Fermenation.kinetic_model` will be used.
    
    Notes
    -----
    Either N or V must be given.
    
    Examples
    --------
    Simulate a Fermentation object which models batch fermentation for the
    production of 1st generation ethanol using yeast.
    
    >>> from biorefineries.cane import create_sugarcane_chemicals
    >>> from biosteam.units import Fermentation
    >>> from biosteam import Stream, settings
    >>> settings.set_thermo(create_sugarcane_chemicals())
    >>> feed = Stream('feed',
    ...               Water=1.20e+05,
    ...               Glucose=1.89e+03,
    ...               Sucrose=2.14e+04,
    ...               DryYeast=1.03e+04,
    ...               units='kg/hr',
    ...               T=32+273.15)
    >>> F1 = NRELFermentation('F1',
    ...                   ins=feed, outs=('CO2', 'product'),
    ...                   tau=8, efficiency=0.90, N=8)
    >>> F1.simulate()
    >>> F1.show()
    NRELFermentation: F1
    ins...
    [0] feed
        phase: 'l', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): Water    6.66e+03
                        Glucose  10.5
                        Sucrose  62.5
                        Yeast    456
    outs...
    [0] CO2
        phase: 'g', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): Water    9.95
                        Ethanol  3.71
                        CO2      244
    [1] product
        phase: 'l', T: 305.15 K, P: 101325 Pa
        flow (kmol/hr): Water    6.59e+03
                        Ethanol  240
                        Glucose  4.07
                        Yeast    532
    >>> F1.results()
    Fermentation                                       Units        F1
    Electricity         Power                             kW      66.6
                        Cost                          USD/hr       5.2
    Chilled water       Duty                           kJ/hr -1.41e+07
                        Flow                         kmol/hr  9.42e+03
                        Cost                          USD/hr      70.3
    Design              Reactor volume                    m3       247
                        Batch time                        hr      12.6
                        Loading time                      hr      1.57
                        Number of reactors                           8
                        Recirculation flow rate        m3/hr      17.7
                        Reactor duty                   kJ/hr -1.41e+07
                        Cleaning and unloading time       hr         3
                        Working volume fraction                    0.9
    Purchase cost       Heat exchangers (x8)             USD  1.57e+05
                        Reactors (x8)                    USD  1.87e+06
                        Agitators (x8)                   USD  1.17e+05
                        Cleaning in place                USD  8.89e+04
                        Recirculation pumps (x8)         USD  1.26e+05
    Total purchase cost                                  USD  2.36e+06
    Utility cost                                      USD/hr      75.5
    
    
    """
    line = 'Fermentation'
    _ins_size_is_fixed = False
    
    #: tuple[float] Kinetic parameters for the kinetic model. Default constants are fitted for Oliveria's model (mu_m1, mu_m2, Ks1, Ks2, Pm1, Pm2, Xm, Y_PS, a)
    kinetic_constants = (0.31,  # mu_m1
                         1.01,  # mu_m2
                         1.88,  # Ks1
                         2.81,  # Ks2
                         82.8,  # Pm1
                         108.2, # Pm2
                         113.4, # Xm
                         0.45,  # Y_PS
                         0.18)  # a
    
    def _init(self, tau, N=None, V=None, T=305.15, P=101325., Nmin=2, Nmax=36,
              efficiency=None, iskinetic=False, fermentation_reaction=None,
              cell_growth_reaction=None):
        NRELBatchBioreactor._init(self, tau=tau, N=N, V=V, T=T, P=P, Nmin=Nmin, Nmax=Nmax)
        self._load_components()
        self.iskinetic = iskinetic
        chemicals = self.chemicals
        self.hydrolysis_reaction = Reaction('Sucrose + Water -> 2Glucose', 'Sucrose', 1.00, chemicals)
        if fermentation_reaction is None:
            if efficiency is None: efficiency = 0.9
            fermentation_reaction = Reaction('Glucose -> 2Ethanol + 2CO2',  'Glucose', efficiency, chemicals)
        else:
            efficiency = fermentation_reaction.X
        self.fermentation_reaction = fermentation_reaction 
        if cell_growth_reaction is None:
            cell_growth_reaction = Reaction('Glucose -> Yeast', 'Glucose', 0.70, chemicals, basis='wt')
            cell_growth_reaction.basis = 'mol'
        self.cell_growth_reaction = cell_growth_reaction
        if all([i in self.chemicals for i in ('FFA', 'DAG', 'TAG', 'Glycerol')]):
            self.lipid_reaction = self.oil_reaction = ParallelReaction([
                Reaction('TAG + 3Water -> 3FFA + Glycerol', 'TAG', 0.23, chemicals),
                Reaction('TAG + Water -> FFA + DAG', 'TAG', 0.02, chemicals)
            ])
        else:
            self.lipid_reaction = None
        self.efficiency = efficiency
        
    def _calc_efficiency(self, feed, tau): # pragma: no cover
        # Get initial concentrations
        IDs = 'Yeast', 'Ethanol', 'Glucose', 
        X0, P0, S0 = feed.imass[IDs] / feed.F_vol
        
        # Integrate to get final concentration
        t = np.linspace(0, tau, 1000)
        C_t = odeint(self.kinetic_model, (X0, P0, S0), t,
                     args=self.kinetic_constants)
        # Cache data
        self._X = C_t[:, 0]
        self._P = C_t[:, 1]
        self._S = S = C_t[:, 2]
        
        # Calculate efficiency
        Sf = S[-1]
        Sf = Sf if Sf > 0 else 0
        Y_PS = self.kinetic_constants[-2]
        eff = (S0 - Sf)/S0 * Y_PS/0.511
        return eff
        
    @staticmethod
    def kinetic_model(z, t, *kinetic_constants): # pragma: no cover
        """
        Return change of yeast, ethanol, and substrate concentration in kg/m3.
        
        Parameters
        ----------
        z : Iterable with (X, E, S) [-]:
            * X: Yeast concentration (kg/m3)
            * P: Ethanol concentration (kg/m3)
            * S: Substrate concentration (kg/m3)
        
        t : float
            Time point
        
        *kinetic_constants
            * mu_m1: Maximum specific growth rate (1/hr)
            * mu_m2: Maximum specific ethanol production rate (g-product/g-cell-hr)
            * Ks1: Sugar saturation constant for growth (g/L)
            * Ks2: Sugar saturation constant for product (g/L)
            * Pm1: Maximum product concentration at zero growth [mu_m1=0] (g/L)
            * Pm2: Maximum product concentration [mu_m2=0] (g/L)
            * Xm: Maximum cell concentration [mu_m1=0] (g/L)
            * Y_PS: Ethanol yield based on sugar consumed
            * a: Toxic power
                
        """
        mu_m1, mu_m2, Ks1, Ks2, Pm1, Pm2, Xm, Y_PS, a = kinetic_constants
        
        # Current yeast, ethanol, and glucose concentration (kg/m3)
        X, P, S = z
        
        # Compute coefficients
        if P > Pm1: P = Pm1
        mu_X = mu_m1 * (S/(Ks1 + S)) * (1 - P/Pm1)**a*((1-X/Xm))
        mu_P = mu_m2 * (S/(Ks2 + S)) * (1 - P/Pm2)
        mu_S = mu_P / 0.45
        
        # Compute derivatives
        dXdt = mu_X * X
        dPdt = (mu_P * X)
        dSdt =  - mu_S * X
        return (dXdt, dPdt, dSdt)

    @property
    def efficiency(self):
        return self.fermentation_reaction.X
    @efficiency.setter
    def efficiency(self, efficiency):
        self.fermentation_reaction.X = efficiency

    def _run(self):
        vent, effluent = self.outs
        effluent.mix_from(self.ins)
        self.hydrolysis_reaction.force_reaction(effluent)
        if self.iskinetic:
            self.fermentation_reaction.X = self._calc_efficiency(effluent, self._tau)
        self.fermentation_reaction.force_reaction(effluent)
        self.cell_growth_reaction.force_reaction(effluent)
        if self.lipid_reaction: self.lipid_reaction.force_reaction(effluent)
        effluent.empty_negative_flows()
        vent.empty()
        vent.receive_vent(effluent, energy_balance=False)

BatchBioreactor = NRELBatchBioreactor
Fermentation = NRELFermentation


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autoclass:: biosteam.units.mixing.Mixer
.. autoclass:: biosteam.units.mixing.SteamMixer
.. autoclass:: biosteam.units.mixing.MockMixer

"""
from .._unit import Unit
from thermosteam._graphics import mixer_graphics
import flexsolve as flx
import biosteam as bst
import numpy as np
from typing import Optional

__all__ = ('Mixer', 'SteamMixer', 'FakeMixer', 'MockMixer')

class Mixer(Unit):
    """
    Create a mixer that mixes any number of streams together.
    
    Parameters
    ----------
    ins : 
        Inlet fluids to be mixed.
    outs : 
        Mixed outlet fluid.
    rigorous :
        Whether to perform vapor-liquid equilibrium.
    
    Notes
    -----
    When streams at different pressures are mixed, BioSTEAM assumes valves 
    reduce the pressure of the streams being mixed to prevent backflow 
    (pressure needs to decrease in the direction of flow according to 
    Bernoulli's principle). The outlet pressure will be the minimum pressure
    of all inlet streams.
    
    Examples
    --------
    Mix two streams:
    
    >>> from biosteam import units, settings, Stream
    >>> settings.set_thermo(['Ethanol', 'Water'], cache=True)
    >>> s1 = Stream('s1', Water=20, T=350)
    >>> s2 = Stream('s2', Ethanol=30, T=300)
    >>> M1 = units.Mixer('M1', ins=(s1, s2), outs='s3')
    >>> M1.simulate()
    >>> M1.show()
    Mixer: M1
    ins...
    [0] s1
        phase: 'l', T: 350 K, P: 101325 Pa
        flow (kmol/hr): Water  20
    [1] s2
        phase: 'l', T: 300 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  30
    outs...
    [0] s3
        phase: 'l', T: 315.14 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  30
                        Water    20
    
    
    """
    _graphics = mixer_graphics
    _N_outs = 1
    _N_ins = 2
    _ins_size_is_fixed = False
    
    Q = 0
    specified_variable = 'Q'
    equation_node_names = (
        'overall_material_balance_node', 
        'energy_balance_node',
    )
    
    def initialize_overall_material_balance_node(self):
        self.overall_material_balance_node.set_equations(
            outputs=[i.F_node for i in self.outs],
            inputs=[j for i in self.ins if (j:=i.F_node)],
        )
        
    def initialize_energy_balance_node(self):
        self.energy_balance_node.set_equations(
            inputs=(
                self.T_node, 
                *[i.F_node for i in (*self.ins, *self.outs)],
                *[j for i in self.ins if (j:=i.E_node)],
            ),
            outputs=[
                j for i in self.outs if (j:=i.E_node)
            ],
        )
    
    @property
    def T_node(self):
        if hasattr(self, '_T_node'): return self._T_node
        self._T_node = var = bst.VariableNode(f"{self.node_tag}.T", lambda: self.outs[0].T)
        return var 
    
    @property
    def B_node(self):
        if hasattr(self, '_B_node'): return self._B_node
        self._B_node = var = bst.VariableNode(f"{self.node_tag}.B", lambda: self.B)
        return var
        
    def get_E_node(self, stream):
        return self.B_node if self.outs[0].phases == ('g', 'l') else self.T_node
    
    def _assert_compatible_property_package(self): 
        pass # Not necessary for mixing streams
    
    def _init(self, rigorous: Optional[bool]=False,
              conserve_phases: Optional[bool]=False):
        self.rigorous = rigorous
        self.conserve_phases = conserve_phases
    
    def _run(self):
        s_out, = self.outs
        s_out.mix_from(self.ins, vle=self.rigorous,
                       conserve_phases=getattr(self, 'conserve_phases', None))
        V = s_out.vapor_fraction
        if V == 0:
            self.B = 0
        elif V == 1:
            self.B = np.inf
        else:
            self.B = V / (1 - V)
    
    def _update_nonlinearities(self): self._run()
    
    def _update_energy_coefficient(self, stream, coefficients):
        if self.outs[0].phases != ('g', 'l'):
            C = stream.C
            coefficients[self, 'T'] = -C
            H_ref = C * stream.T
        else:
            H_ref = 0
        return H_ref
    
    def _create_energy_balance_equations(self):
        fresh_inlets, process_inlets, equations = self._begin_energy_equations()
        Q = self.Q + sum([i.H for i in fresh_inlets])
        outlet = self.outs[0]
        if outlet == ('g', 'l'):
            coeff = {}
            for i in self.flat_outs: coeff[i, 'F_mol'] = i.h
        else:
            coeff = {(outlet, 'F_mol'): outlet.h}
            coeff[self, 'T'] = C = outlet.C
            Q += C * outlet.T
        for i in process_inlets: 
            if i.isempty(): continue
            coeff[i, 'F_mol'] = -i.h
            Q -= i._update_energy_coefficient(coeff)
        return [(coeff, Q)]
    
    def _create_bulk_balance_equations(self):
        fresh_inlets, process_inlets, equations = self._begin_bulk_equations()
        eq_overall = {}
        outlet = self.outs[0]
        eq_overall[outlet, 'F_mol'] = 1
        for i in process_inlets:
            if i in eq_overall: del eq_overall[i, 'F_mol']
            else: eq_overall[i, 'F_mol'] = -1
        equations.append(
            (eq_overall, sum([i.F_mol for i in fresh_inlets]))
        )
        if outlet.phases == ('g', 'l'):
            top, bottom = self.flat_outs
            eq_outs = {}
            eq_outs[top, 'F_mol'] = 1
            eq_outs[bottom, 'F_mol'] = -top.F_mol / bottom.F_mol
            equations.append(
                (eq_outs, 0)
            )
        return equations
    
    def _reset_bulk_variable(self):
        if self.outs[0].phases == ('g', 'l'):
            partition = self.partition
            top, bottom = self.partition.outs
            IDs = partition.IDs
            partition.B = top.imol[IDs].sum() / bottom.imol[IDs].sum()
    
    def _create_material_balance_equations(self, composition_sensitive):
        fresh_inlets, process_inlets, equations = self._begin_material_equations(composition_sensitive)
        outlet, = self.outs
        if len(outlet) == 1:
            ones = np.ones(self.chemicals.size)
            minus_ones = -ones
            zeros = np.zeros(self.chemicals.size)
            
            # Overall flows
            eq_overall = {outlet: ones}
            for i in process_inlets: eq_overall[i] = minus_ones
            equations.append(
                (eq_overall, sum([i.mol for i in fresh_inlets], zeros))
            )
        else:
            top, bottom = outlet
            ones = np.ones(self.chemicals.size)
            minus_ones = -ones
            zeros = np.zeros(self.chemicals.size)
            
            # Overall flows
            eq_overall = {}
            for i in outlet: 
                eq_overall[i] = ones
            for i in process_inlets:
                eq_overall[i] = minus_ones
            equations.append(
                (eq_overall, sum([i.mol for i in fresh_inlets], zeros))
            )
            
            # Top to bottom flows
            B = self.B
            eq_outs = {}
            if B == np.inf:
                eq_outs[bottom] = ones
            elif B == 0:
                eq_outs[top] = ones
            else:
                bp = outlet.bubble_point_at_P()
                outlet.T = bp.T
                S = bp.K * B
                eq_outs[top] = ones
                eq_outs[bottom] = -S
            equations.append(
                (eq_outs, zeros)
            )
        return equations
    
    def _update_variable(self, variable, value):
        if variable == 'T':
            self.outs[0].T = value
        else:
            raise ValueError('cannot update variable')


class SteamMixer(Unit):
    """
    Create a mixer that varies the flow of steam to achieve a specified outlet
    pressure and varies the flow of process water to achieve a specified 
    solids loading (by wt).
    
    Parameters
    ----------
    ins : 
        * [0] Feed    
        * [1] Steam
        * [2] Process water
    outs : 
        Mixed product.
    P : float
        Outlet pressure.
    T : float
        Outlet temperature.
    solids_loading : float, optional
        Final solids loading after mixing in process water.
    soilds_loading_includes_steam : bool, optional
        Whether to include steam in solids loading calculation.
    
    Examples
    --------
    >>> import biosteam as bst
    >>> bst.settings.set_thermo(['Water', 'Glucose'])
    >>> feed = bst.Stream('feed', Water=10, Glucose=10)
    >>> M1 = bst.SteamMixer(None, ins=[feed, 'steam', 'process_water'], outs='outlet', T=431.15, P=557287.5, solids_loading=0.3)
    >>> M1.simulate()
    >>> M1.show('cwt100') # Note that outlet solids loading is not exactly 0.3 because of the steam requirement.
    SteamMixer
    ins...
    [0] feed
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water    9.09
                         Glucose  90.9
                         -------  1.98e+03 kg/hr
    [1] steam
        phase: 'g', T: 454.77 K, P: 1.041e+06 Pa
        composition (%): Water  100
                         -----  1.28e+03 kg/hr
    [2] process_water
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water  100
                         -----  4.02e+03 kg/hr
    outs...
    [0] outlet
        phase: 'l', T: 431.15 K, P: 557288 Pa
        composition (%): Water    75.3
                         Glucose  24.7
                         -------  7.29e+03 kg/hr
    
    >>> M1.solids_loading_includes_steam = True
    >>> M1.simulate()
    >>> M1.show('cwt100') # Now the outlet solids content is exactly 0.3
    SteamMixer
    ins...
    [0] feed
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water    9.09
                         Glucose  90.9
                         -------  1.98e+03 kg/hr
    [1] steam
        phase: 'g', T: 454.77 K, P: 1.041e+06 Pa
        composition (%): Water  100
                         -----  1.02e+03 kg/hr
    [2] process_water
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water  100
                         -----  3.01e+03 kg/hr
    outs...
    [0] outlet
        phase: 'l', T: 431.15 K, P: 557288 Pa
        composition (%): Water    70
                         Glucose  30
                         -------  6.01e+03 kg/hr
    
    """
    _N_outs = 1
    _N_ins = 3
    _ins_size_is_fixed = False
    _graphics = mixer_graphics
    installation_cost = purchase_cost = 0.
    def _init(self, P, T=None, solids_loading=None, 
             liquid_IDs=['7732-18-5'], solid_IDs=None,
             solids_loading_includes_steam=None):
        self.P = P
        self.T = T
        self.solids_loading = solids_loading
        self.solids_loading_includes_steam = solids_loading_includes_steam
        self.liquid_IDs = tuple(liquid_IDs)
        self.solid_IDs = solid_IDs
    
    @property
    def steam(self):
        return self.ins[1]
    
    def reset_cache(self, isdynamic=None): 
        for utility in bst.HeatUtility.heating_agents:
            if utility.P > self.P: break
        self.steam.copy_like(utility)
    
    def pressure_objective_function(self, steam_mol):
        mixed = self.outs[0]
        self.steam.imol['7732-18-5'] = steam_mol # Only change water
        if self.P: mixed.P = self.P # Assume pumps take care of this
        if self.solids_loading_includes_steam and self.solids_loading:
            solids_loading = self.solids_loading
            feed, steam, process_water, *others = self.ins
            process_water.empty()
            feeds = self.ins
            chemicals = self.chemicals
            index = chemicals.get_index(self.liquid_IDs)
            available_water = 18.01528 * sum([(j.sum() if hasattr((j:=i.mol[index]), 'sum') else j) for i in feeds if i])
            solid_IDs = self.solid_IDs
            if solid_IDs:
                F_mass_solids = sum([i.imass[solid_IDs].sum() for i in feeds if i])
            else:
                F_mass_feed = sum([i.F_mass for i in feeds if i])
                F_mass_solids = F_mass_feed - available_water
            required_water = F_mass_solids * (1. - solids_loading) / solids_loading
            process_water.imol['7732-18-5'] = max(required_water - available_water, 0.) / 18.01528
        
        mixed.mix_from(self.ins, energy_balance=False)
        H = sum([i.H for i in self.ins])
        Tmax = mixed.chemicals.Water.Tc - 1
        mixed.T = Tmax
        Hmax = mixed.H
        if H > Hmax:
            mixed.T = Tmax + (H - Hmax) / mixed.chemicals.Water.Cn('l', Tmax)
        else:
            mixed.H = H
        if self.T:
            return self.T - mixed.T
        else: # If no pressure, assume it is at the boiling point
            P_new = mixed.chemicals.Water.Psat(min(mixed.T, Tmax))
            return self.P - P_new
    
    def _setup(self):
        super()._setup()
        if self.steam.isempty(): self.reset_cache()
    
    def _run(self):
        solids_loading = self.solids_loading
        if solids_loading is not None and not self.solids_loading_includes_steam:
            # Solids loading need to be achieved first before mixing with steam
            # to avoid pumping issues (see Humbird 2011 NREL report).
            feed, steam, process_water, *others = self.ins
            process_water.empty()
            feeds = [feed, *others]
            chemicals = self.chemicals
            index = chemicals.get_index(self.liquid_IDs)
            available_water = 18.01528 * sum([(j.sum() if hasattr((j:=i.mol[index]), 'sum') else j) for i in feeds if i])
            solid_IDs = self.solid_IDs
            if solid_IDs:
                F_mass_solids = sum([i.imass[solid_IDs].sum() for i in feeds if i])
            else:
                F_mass_feed = sum([i.F_mass for i in feeds if i])
                F_mass_solids = F_mass_feed - available_water
            required_water = F_mass_solids * (1. - solids_loading) / solids_loading
            process_water.imol['7732-18-5'] = max(required_water - available_water, 0.) / 18.01528
        else:
            steam = self.steam
        steam_mol = steam.F_mol or 1.
        f = self.pressure_objective_function
        steam_mol = flx.IQ_interpolation(f, *flx.find_bracket(f, 0., steam_mol, None, None), 
                                         xtol=1e-2, ytol=1e-4, 
                                         maxiter=500, checkroot=False)
        self.outs[0].P = self.P
        
    def _design(self): 
        steam = self.ins[1]
        warm_process_water = self.ins[2]
        mixed = self.outs[0]
        self.add_heat_utility(steam.H + warm_process_water.H, mixed.T)
        
class MockMixer(Unit):
    """
    Create a MockMixer object that does nothing when simulated.
    """
    _graphics = Mixer._graphics
    _N_ins = 2
    _N_outs = 1
    _ins_size_is_fixed = False
    
    def _run(self): pass

MockMixer.line = 'Mixer'
FakeMixer = MockMixer    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from numba import njit
import numpy as np
from thermosteam import Stream
from warnings import warn
from typing import Optional, Callable
from scipy.spatial.distance import cdist
from itertools import combinations, product
from ._parameter import Parameter
from .._system import System, JointRecycleData

__all__ = (
    'GenericResponse',
    'ConvergenceModel',
    'NullConvergenceModel',
    'Average',
    'LinearRegressor',
    'InterceptLinearRegressor',
)

@njit(cache=True)
def R2(y_actual, y_predicted):
    dy = y_actual - y_predicted
    dy_m = y_actual - np.mean(y_actual)
    SSR = np.dot(dy, dy)
    SST = np.dot(dy_m, dy_m)
    if SST == 0.:
        if SSR == 0: 
            return 1.
        else:
            return -np.inf
    else:
        return 1 - SSR / SST

@njit(cache=True)
def fit_linear_model(X, y):
    Xt = X.transpose()
    return np.linalg.inv(Xt @ X) @ Xt @ y

class LinearRegressor:
    __slots__ = (
        'coefficients',
    )
    def __init__(self):
        self.coefficients = None
        
    def fit(self, X, y):
        self.coefficients = fit_linear_model(X, y)
    
    def predict(self, x):
        return np.dot(self.coefficients, x[0])

    def __repr__(self):
        return f"{type(self).__name__}()"


class InterceptLinearRegressor:
    __slots__ = (
        'coefficients',
    )
    def __init__(self):
        self.coefficients = None
        
    def fit(self, X, y):
        m, n = X.shape
        Xi = np.ones([m, n + 1])
        Xi[:, 1:] = X 
        self.coefficients = fit_linear_model(Xi, y)
    
    def predict(self, x):
        xi = np.ones(x.shape[1] + 1)
        xi[1:] = x[0]
        return np.dot(self.coefficients, xi)

    def __repr__(self):
        return f"{type(self).__name__}()"


class Average:
    __slots__ = (
        'mean',
    )
    def __init__(self):
        self.mean = None
        
    def fit(self, X, y):
        self.mean = np.mean(y)
    
    def predict(self, x):
        return self.mean

    def __repr__(self):
        return f"{type(self).__name__}()"
    
    
fast_fit_model_types = set([
    LinearRegressor, 
    InterceptLinearRegressor,
    Average
])
linear_model_types = set([
    LinearRegressor, 
    InterceptLinearRegressor, 
])
    
def recycle_response(recycle, name, model):
    if name == 'T':
        response = RecycleTemperature(recycle, model)
    elif name == 'P':
        response = RecyclePressure(recycle, model)
    else:
        response = RecycleFlow(recycle, name, model)
    return response
    

class Response:
    __slots__ = (
        'element', 'model', 'predictors',
        'max', 'min',
    )
    def __init__(self, element, model):
        self.element = element
        self.model = model
        self.predictors = [] # Indices of predictors
        self.max = self.min = None
    
    def __hash__(self):
        return hash((self.element, self.name))
    
    def __eq__(self, other):
        return (self.element, self.name) == other
    
    def fit(self, X, y):
        self.model.fit(X[:, self.predictors], y)

    def predict(self, x):
        return self.filter_value(
            float(self.model.predict(x[None, self.predictors]))
        )
    
    def predict_locally(
            self, x, X, y, distance, weight
        ):
        index = self.predictors
        x = x[None, index]
        X = X[:, index]
        model = self.model
        distances = cdist(x, X, metric=distance)
        exact_match = distances == 0
        if exact_match.any(): return y[exact_match[0]].mean()
        w = weight(distances)
        X = X * w.transpose()
        y = y * w[0]
        self.model.fit(X, y)
        yi = model.predict(x)
        return self.filter_value(float(yi))
    
    def update_limits(self, value):
        if self.max is None:
            self.max = value
            self.min = value
        elif value > self.max:
            self.max = value
        elif value < self.min:
            self.min = value
    
    def filter_value(self, value):
        if value < self.min:
            return self.min
        elif value > self.max:
            return self.max
        else:
            return value
    
    def __str__(self):
        return f"{self.element}.{self.name}"
    
    def __repr__(self):
        return f"{type(self).__name__}({self.element!r})"


class GenericResponse(Response):
    __slots__ = (
        'name',
        'units',
    )
    
    def __init__(self, 
            element, name, model=None, units=None, max=None, min=None,
        ):
        self.name = name
        self.units = units
        self.max = max
        self.min = min
        super().__init__(element, model)
        
    def get(self):
        return getattr(self.element, self.name)
    
    def set(self, value):
        setattr(self.element, self.name, self.filter_value(value))
    
    def __repr__(self):
        return f"{type(self).__name__}({self.element!r}, {self.name!r})"
        

class RecycleTemperature(Response):
    __slots__ = (
        'element',
    )
    name = 'T'
    units = 'K'
    
    def get(self):
        return self.element.T
    
    def set(self, value):
        self.element.T = value
    

class RecyclePressure(Response):
    __slots__ = (
        'element',
    )
    name = 'P'
    units = 'Pa'
    
    def set(self, value):
        self.element.P = value
        
    def get(self):
        return self.element.P


class RecycleFlow(Response):
    __slots__ = (
        'element',
        'name',
    )
    units = 'kmol/hr'
    
    def __init__(self, element, name, model):
        self.name = name
        super().__init__(element, model)
        
    def set(self, value):
        self.element.imol[self.name] = self.filter_value(value)
    
    def get(self):
        return self.element.imol[self.name]

    def __repr__(self):
        return f"{type(self).__name__}({self.element!r}, {self.name!r})"
    

class ConvergenceModel:
    __slots__ = (
        'system',
        'predictors', 
        'responses', 
        'data',
        'case_study',
        'model_type',
        'recess',
        'distance',
        'fitted',
        'sample_length',
        'sample_min',
        'sample_range',
        'weight',
        'nfits',
        'local_weighted',
        'normalization',
        'interaction_pairs',
        'predictor_index',
        'save_prediction',
    )
    absolute_response_tolerance = 0.001
    relative_response_tolerance = 0.01
    def __init__(self, 
            predictors: tuple[Parameter],
            model_type: Optional[Callable]=None,
            recess: Optional[int]=None, 
            distance: Optional[str]=None,
            weight:  Optional[Callable]=None,
            nfits: Optional[int]=None,
            local_weighted: Optional[int]=None,
            system: Optional[System] = None,
            responses: Optional[set[Response]]=None,
            interaction_pairs: Optional[bool] = None,
            normalization: Optional[bool] = None,
            load_responses: Optional[bool] = None,
            save_prediction: Optional[bool] = True,
        ):
        if system is None:
            systems = set([i.system for i in predictors])
            try:
                system, = systems
            except:
                if systems:
                    raise ValueError('predictors must share the same system')
                else:
                    raise ValueError('no system available')
        self.system = system
        self.responses = set() if responses is None else responses
        if model_type is None: 
            if len(predictors) > 10:
                model_type = LinearRegressor
            else:
                model_type = InterceptLinearRegressor
        if isinstance(model_type, str):
            model_type = model_type.lower()
            if model_type == 'linear regressor':
                model_type = LinearRegressor
            elif model_type == 'intercept linear regressor':
                model_type = InterceptLinearRegressor
            elif model_type == 'linear svr': # linear support vector machine regression
                from sklearn.svm import LinearSVR
                from sklearn.pipeline import make_pipeline
                from sklearn.preprocessing import StandardScaler
                model_type = lambda: make_pipeline(
                    StandardScaler(),
                    LinearSVR()
                )
                if nfits is None: nfits = 100
            elif model_type == 'svr':
                from sklearn.svm import SVR
                from sklearn.pipeline import make_pipeline
                from sklearn.preprocessing import StandardScaler
                model_type = lambda: make_pipeline(
                    StandardScaler(),
                    SVR()
                )
                if nfits is None: nfits = 10
            elif model_type == 'average':
                model_type = Average
            else:
                raise ValueError('unknown model type {model_type!r}')
        if recess is None: 
            if model_type in fast_fit_model_types:
                recess = 0
            else:
                recess = 5 * sum([i.coupled for i in predictors])
        if interaction_pairs is None: interaction_pairs = False
        if local_weighted is None:
            if model_type in fast_fit_model_types:
                local_weighted = True
            else:
                local_weighted = False
        if local_weighted:
            if normalization is None: normalization = model_type in linear_model_types
            if distance is None: distance = 'cityblock'
            if recess: raise ValueError('local weighted recycle model cannot recess')
            if nfits: raise ValueError('local weighted recycle model must fit every time; cannot pass nfits argument')
            if weight is None: weight = lambda x: np.exp(-x / 2)
        self.model_type = model_type
        self.recess = recess
        self.distance = distance
        self.fitted = 0
        self.weight = weight
        self.nfits = nfits
        self.local_weighted = local_weighted
        self.interaction_pairs = interaction_pairs
        self.normalization = normalization
        self.load_predictors(predictors)
        self.save_prediction = save_prediction
        if load_responses is None or load_responses: self.load_responses()
        
    def with_interactions(self, sample):
        n = len(sample)
        new_sample = np.zeros(self.sample_length)
        new_sample[:n] = sample
        for i, (a, b) in self.interaction_pairs:
            new_sample[i] = sample[a] * sample[b]
        return new_sample
    
    def normalize_sample(self, sample):
        return (sample - self.sample_min) / self.sample_range
    
    def reframe_sample(self, sample, predictors=None):
        if predictors is not None and predictors != self.predictors:
            index = {j: i for i, j in enumerate(self.predictors)}
            sample = self.data['samples'][-1]
            for p, s in zip(predictors, sample): sample[index[p]] = s
        if self.predictor_index is not None:
            sample = np.asarray(sample)[self.predictor_index]
        if self.normalization:
            sample = self.normalize_sample(sample)
        if self.interaction_pairs:
            sample = self.with_interactions(sample)
        return sample
        
    def fitted_responses(self):
        data = self.data
        responses = self.responses
        fitted = {i: [] for i in responses}
        samples = np.array(data['samples'])
        try: self.fit()
        except: return {}
        for i, sample in enumerate(samples):
            for response in responses:
                fitted[response].append(
                    response.predict(sample)
                )
        return fitted
    
    def R2(self, last=None):
        predicted, predicted_dct = self.R2_predicted(last)
        fitted, fitted_dct = self.R2_fitted()
        return (
            {'predicted': predicted,
             'fitted': fitted},
            {'predicted': predicted_dct,
             'fitted': fitted_dct},   
        )
            
    def _R2(self, dataset, last=None):
        results = {}
        data = self.data
        actual = data['actual']
        fitted = dataset == 'fitted'
        predicted = self.fitted_responses() if fitted else data[dataset]
        for response in self.responses:
            name = str(response)
            y_actual = np.array(actual[response])
            y_predicted = np.array(predicted[response])
            if last is None: 
                index = len(y_predicted)
            else:
                index = last
            y_actual = y_actual[-index:]
            y_predicted = y_predicted[-index:]
            results[name] = R2(
                y_actual,
                y_predicted,
            )
        mean = sum(results.values()) / len(results)
        lb = min(results.values())
        ub = max(results.values())
        return {'min': lb, 'mean': mean, 'max': ub}, results
        
    def R2_predicted(self, last=None):
        return self._R2('predicted', last)
    
    def R2_fitted(self):
        return self._R2('fitted')
    
    def practice(self, case_study, predictors=None):
        """
        Predict and set recycle responses given the sample, then append actual
        simulation result to data.
        
        Must be used in a with-statement as follows:
            
        ```python
        with recycle_model.practice(case_study): # the case study is an unsimulated sample
            recycle_model.system.simulate() # Or other simulation code.
            
        ```
        
        This method effectively does the same as running:
            
        ```python
        recycle_model.predict(sample)
        recycle_data = recycle_model.evaluate_system_convergence() # Or other simulation code.
        recycle_model.append_data(sample, recycle_data)
        ```
        
        Warning
        -------
        Does nothing if not used in with statement containing simulation.
        
        """
        self.case_study = self.reframe_sample(case_study, predictors)
        return self
        
    def __enter__(self):
        data = self.data
        if self.save_prediction: predicted = data['predicted']
        actual = data['actual']
        sample_list = data['samples']
        n_samples = len(sample_list)
        samples = np.array(sample_list)
        case_study = self.case_study
        if self.save_prediction: 
            if self.local_weighted:
                for response in self.responses:
                    prediction = response.predict_locally(
                        case_study, samples, np.array(actual[response]),
                        self.distance, self.weight,
                    )
                    response.set(prediction)
                    predicted[response].append(prediction)
            elif (not n_samples % (self.recess + 1)  # Recess is over
                  and (self.nfits is None or self.fitted < self.nfits)):
                self.fitted += 1
                for response in self.responses:
                    response.fit(samples, np.array(actual[response]))
                    prediction = response.predict(case_study) 
                    response.set(prediction)
                    predicted[response].append(prediction)
            elif self.fitted:
                for response in self.responses:
                    prediction = response.predict(case_study) 
                    response.set(prediction)
                    predicted[response].append(prediction)
        elif self.local_weighted:
            for response in self.responses:
                response.set(
                    response.predict_locally(
                        case_study, samples, np.array(actual[response]),
                        self.distance, self.weight,
                    )
                )
        elif (not n_samples % (self.recess + 1)  # Recess is over
              and (self.nfits is None or self.fitted < self.nfits)):
            self.fitted += 1
            for response in self.responses:
                response.fit(samples, np.array(actual[response]))
                response.set(response.predict(case_study))
        elif self.fitted:
            for response in self.responses:
                response.set(response.predict(case_study))
        sample_list.append(case_study)
    
    def __exit__(self, type, exception, traceback, total=[]):
        del self.case_study
        data = self.data
        if exception and self.fitted:
            del data['samples'][-1]
            raise exception
        actual = data['actual']
        for response in self.responses:
            value = response.get()
            response.update_limits(value)
            actual[response].append(value)
        
    def evaluate_system_convergence(self, sample, default=None, **kwargs):
        system = self.system
        for p, value in zip(self.predictors, sample): 
            p.setter(value)
            p.last_value = value
        try:
            system.simulate(design_and_cost=False, **kwargs)
        except Exception as error:
            if default is None: raise error
            # TODO: py3.11, add eror context
            #     raise RuntimeError(
            #         'baseline simulation failed; '
            #         'could not load responses for convergence prediction model'
            #     )
            system.empty_recycles()
            recycles_data = default
        else:
            recycles_data = system.get_recycle_data()
        return recycles_data
       
    def load_predictors(self, predictors):
        predictor_index = np.array([i.coupled for i in predictors])
        self.predictor_index = predictor_index = None if predictor_index.all() else np.where(predictor_index)[0]
        if predictor_index is not None: predictors = [predictors[i] for i in predictor_index]
        self.predictors = predictors
        bounds = [i.bounds for i in predictors]
        n = len(predictors)
        if self.interaction_pairs:
            self.interaction_pairs = [*enumerate(combinations(range(n), 2), n)]
            self.sample_length = int(n * (n - 1) / 2) + n
        else:
            self.interaction_pairs = ()
        if self.normalization:
            self.sample_min = sample_min = np.zeros(n)
            self.sample_range = sample_range = np.zeros(n)
            for i, (lb, ub) in enumerate(bounds):
                sample_min[i] = lb
                sample_range[i] = ub - lb
    
    def load_responses(self): 
        """
        Select material responses and their respective predictors through single point 
        sensitivity. Also store the simulation data for fitting later.
        """
        predictors = self.predictors
        responses = self.responses
        hooks = [i.hook for i in predictors]
        all_bounds = [i.bounds for i in predictors]
        sample = [i.baseline for i in predictors]
        evaluate = self.evaluate_system_convergence        
        baseline_1 = evaluate(sample)
        values = []
        values_at_bounds = []
        samples = [sample]
        for i, p in enumerate(predictors):
            hook = hooks[i]
            bounds = all_bounds[i]
            for limit in bounds:
                new_sample = sample.copy()
                if hook is not None: limit = hook(limit)
                new_sample[i] = limit
                samples.append(new_sample)
                values.append(
                    evaluate(new_sample, default=baseline_1, recycle_data=baseline_1)
                )
            values_at_bounds.append(values[-2:])
        if self.interaction_pairs:
            # [(lb, lb), (lb, ub), (ub, lb), (ub, ub)]
            pairwise_bounds_index = [*enumerate(product(range(2), range(2)))]
            for _, pair in self.interaction_pairs:
                new_samples = [sample.copy() for i in range(4)]
                for i, bounds_index in pairwise_bounds_index:
                    for b, p in zip(bounds_index, pair):
                        hook = hooks[p]
                        bounds = all_bounds[p]
                        limit = bounds[b]
                        new_sample = new_samples[i]
                        if hook is not None: limit = hook(limit)
                        new_sample[p] = limit
                for new_sample in new_samples:
                    values.append(
                        evaluate(new_sample, default=baseline_1, recycle_data=baseline_1)
                    ) 
                    samples.append(new_sample)
        baseline_2 = evaluate(sample, recycle_data=baseline_1)
        arr1 = baseline_1.to_array()
        arr2 = baseline_2.to_array()
        error = np.abs(arr1 - arr2)
        index, = np.where(error > self.absolute_response_tolerance)
        error = error[index]
        relative_error = error / np.maximum.reduce([np.abs(arr1[index]), np.abs(arr2[index])])
        rtol = self.relative_response_tolerance
        bad_index = [i for i, bad in enumerate(relative_error > rtol) if bad]
        if bad_index:
            keys = baseline_1.get_keys()
            names = baseline_1.get_names()
            relative_error = relative_error[bad_index]
            bad_index = [index[i] for i in bad_index]
            bad_keys = set([keys[i] for i in bad_index])
            bad_names = [names[i] for i in bad_index]
            bad_names = ', '.join(bad_names)
            warn(
                f"inconsistent model; recycle loops on [{bad_names}] do not "
                "match at baseline before and after single point "
                f"sensitivity analysis ({100 * relative_error} % error)",
                RuntimeWarning
            )
        else:
            bad_keys = set()
        self.add_sensitive_reponses(
            baseline_1, values_at_bounds, bad_keys
        )
        self.data = data = {'samples': []}
        for name in ('actual', 'predicted'): 
            data[name] = {key: [] for key in responses}
        predictor_index = self.predictor_index
        self.predictor_index = None
        self.extend_data(samples, values)
        self.predictor_index = predictor_index
        
    def add_sensitive_reponses(self, 
            baseline: JointRecycleData,
            bounds: tuple[JointRecycleData],
            exclude_responses: set[tuple[Stream, str]],
        ):
        model_type = self.model_type
        responses = self.responses
        baseline = baseline.to_dict()
        responses_dct = {}
        rtol = self.relative_response_tolerance
        for p, (lb, ub) in enumerate(bounds):
            if lb is baseline is ub: continue
            recycle_data = (lb.to_dict(), baseline, ub.to_dict())
            keys = set()
            for i in recycle_data: keys.update(i)
            for key in keys:
                if key in exclude_responses: continue
                values = [dct.get(key, 0.) for dct in recycle_data]
                mean = np.mean(values)
                if any([(i - mean) > mean * rtol for i in values]):
                    if key in responses_dct:
                        response = responses_dct[key]
                    else:
                        if isinstance(key, Response):
                            response = key
                        else:
                            response = recycle_response(*key, model_type())
                        responses_dct[key] = response
                        responses.add(response)
                    response.predictors.append(p)
        interaction_pairs = self.interaction_pairs
        for response in responses:
            if interaction_pairs:
                predictors_set = set(response.predictors)
                response.predictors.extend(
                    [i for i, (a, b) in interaction_pairs
                     if a in predictors_set and b in predictors_set]
                )
            if response.model is None: 
                response.model = model_type()
        
    def append_data(self, sample, recycle_data=None):
        data = self.data
        actual = data['actual']
        data['samples'].append(self.reframe_sample(sample))
        dct = recycle_data.to_dict()
        for response in self.responses:
            value = dct.get(response, 0.)
            response.update_limits(value)
            actual[response].append(value)
            
    def extend_data(self, samples, recycle_data):
        for args in zip(samples, recycle_data): self.append_data(*args)
    
    def fit(self):
        data = self.data
        actual = data['actual']
        samples = np.array(data['samples'])
        for response in self.responses:
            response.fit(samples, np.array(actual[response]))    
    
    def predict(self, sample):
        for response in self.responses:
            prediction = response.predict(sample)
            response.set(prediction)
            

class NullConvergenceModel:
    __slots__ = (
        'system',
        'predictors', 
        'predictor_index',
        'responses', 
        'data',
        'case_study',
        'interaction_pairs',
        'normalization',
    )
    absolute_response_tolerance = ConvergenceModel.absolute_response_tolerance
    relative_response_tolerance = ConvergenceModel.relative_response_tolerance
    practice = ConvergenceModel.practice
    evaluate_system_convergence = ConvergenceModel.evaluate_system_convergence
    load_responses = ConvergenceModel.load_responses
    load_predictors = ConvergenceModel.load_predictors
    add_sensitive_reponses = ConvergenceModel.add_sensitive_reponses
    append_data = ConvergenceModel.append_data
    extend_data = ConvergenceModel.extend_data
    reframe_sample = ConvergenceModel.reframe_sample
    
    def __init__(self, 
            predictors: tuple[Parameter],
            system: Optional[System] = None,
            responses: Optional[set[Response]]=None,
            load_responses: Optional[bool] = None,
        ):
        if system is None:
            systems = set([i.system for i in predictors])
            try:
                system, = systems
            except:
                if systems:
                    raise ValueError('predictors do not share the same system')
                else:
                    raise ValueError('no system available')
        self.system = system
        self.interaction_pairs = None
        self.normalization = None
        self.load_predictors(predictors)
        self.responses = set() if responses is None else responses
        if load_responses is None or load_responses: self.load_responses()
    
    def model_type(self): return None
    
    def R2(self, last=None):
        results = {}
        data = self.data
        actual = data['actual']
        null_responses = data['predicted']
        for response in self.responses:
            name = str(response)
            y_actual = np.array(actual[response]) 
            y_predicted = np.array(null_responses[response])
            if last is None: last = len(y_predicted)
            y_actual = y_actual[-last:]
            y_predicted = y_predicted[-last:]
            results[name] = R2(
                y_actual,
                y_predicted,
            )
        mean = sum(results.values()) / len(results)
        lb = min(results.values())
        ub = max(results.values())
        return {'min': lb, 'mean': mean, 'max': ub}, results
        
    def __enter__(self):
        data = self.data
        null_responses = data['predicted']
        sample_list = data['samples']
        sample_list.append(self.case_study)
        for response in self.responses:
            null_responses[response].append(response.get())
    
    def __exit__(self, type, exception, traceback, total=[]):
        del self.case_study
        data = self.data
        if exception: 
            data = self.data
            if exception:
                del data['samples'][-1]
                null_responses = data['predicted']
                for response in self.responses:
                    del null_responses[response][-1]
                raise exception
        actual = data['actual']
        for response in self.responses:
            actual[response].append(response.get())
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import biosteam as bst
from numpy.testing import assert_allclose
from thermo import SRK, PR


def test_isentropic_hydrogen_compressor():
    bst.settings.set_thermo(["H2"])
    bst.settings.chemicals.H2.V.g.method_P = 'IDEAL'
    eta = 0.7
    feed = bst.Stream(H2=1, T=25 + 273.15, P=101325, phase='g')
    K = bst.units.IsentropicCompressor(ins=feed, P=50e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, 1151.9888331779612, 5000000.0],
        rtol=1e-3,
    )
    # check compressor design
    assert K.design_results["Type"] == "Centrifugal"
    assert K.design_results["Driver"] == "Steam turbine"
    ideal_power = 4.920933056717339
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Steam turbine"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.consumption
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
         K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, 0, 1],
        rtol=1e-3,
    )
    pass


def test_isentropic_two_phase_steam_compressor():
    bst.settings.set_thermo(["H2O"])
    bst.settings.chemicals.H2O.V.g.method_P = 'IDEAL'
    # feed is steam-water-mixture
    feed = bst.MultiStream(T=372.75, P=1e5, l=[('H2O', 0.1)], g=[('H2O', 0.9)])
    # from compress until superheated steam (test case for vle parameter)
    eta = 1
    K = bst.units.IsentropicCompressor(ins=feed, P=100e5, eta=eta, vle=True)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, 798.6308938602593, 10000000.0],
        rtol=1e-3,
    )
    # check compressor design
    assert K.design_results["Type"] == "Centrifugal"
    assert K.design_results["Driver"] == "Steam turbine"
    ideal_power = 5.4152214427108305
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Steam turbine"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.consumption
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, 0, 1],
        rtol=1e-3,
    )
    pass


def test_isothermal_hydrogen_compressor():
    thermo = bst.Thermo([bst.Chemical('H2', eos=SRK)])
    thermo.chemicals.H2.V.g.method_P = 'IDEAL'
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream(H2=1, T=298.15, P=20e5, phase='g')

    ideal_power = 2.102711178815804
    ideal_duty = -7260.634742550824

    # eta = 1
    eta = 1
    K = bst.units.IsothermalCompressor(ins=feed, P=350e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, feed.T, 350e5],
        rtol=1e-3,
    )
    # check compressor design
    assert K.design_results["Type"] == "Reciprocating"
    assert K.design_results["Driver"] == "Electric motor"
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Electric motor"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, ideal_duty, 1],
        rtol=1e-3,
    )
    # check heat utility
    expected_duty = ideal_duty / eta
    assert_allclose(
        [K.heat_utilities[0].unit_duty, K.heat_utilities[0].duty, K.heat_utilities[0].flow],
        [expected_duty, expected_duty, 7.5262772846646],
        rtol=1e-3,
    )

    # repeat with eta=0.7
    eta = 0.7
    K = bst.units.IsothermalCompressor(ins=feed, P=350e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, feed.T, 350e5],
        rtol=1e-3,
    )
    # check compressor design
    assert K.design_results["Type"] == "Reciprocating"
    assert K.design_results["Driver"] == "Electric motor"
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Electric motor"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, ideal_duty, 1],
        rtol=1e-3,
    )
    # check heat utility
    expected_duty = ideal_duty / eta
    hu = K.heat_utilities[0]
    assert_allclose(
        [hu.unit_duty, hu.duty, hu.flow],
        [expected_duty, expected_duty, 7.526277284664599 / eta],
        rtol=1e-3,
    )
    pass

def test_polytropic_hydrogen_compressor():
    thermo = bst.Thermo([bst.Chemical('H2', eos=PR)])
    thermo.chemicals.H2.V.g.method_P = 'IDEAL'
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream(H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    # eta=1 outlet state should be identical to ideal isentropic compression
    # schultz method
    eta = 1.0
    K = bst.units.PolytropicCompressor(ins=feed, P=P, eta=eta, method="schultz")
    K.simulate()
    out_poly = K.outs[0]
    K_isen = bst.units.IsentropicCompressor(ins=feed, P=P, eta=eta)
    K_isen.simulate()
    out_isen = K_isen.outs[0]
    assert len(K.heat_utilities) == 0 and K.net_duty == 0.
    assert_allclose(
        [out_poly.T, out_poly.P, out_poly.H, K.power_utility.rate],
        [out_isen.T, out_isen.P, out_isen.H, K_isen.power_utility.rate],
        rtol=1e-3,
    )
    # eta=1, hundseid method
    K = bst.units.PolytropicCompressor(ins=feed, P=P, eta=eta, method="hundseid")
    K.simulate()
    out_poly = K.outs[0]
    assert len(K.heat_utilities) == 0 and K.net_duty == 0.
    assert_allclose(
        [out_poly.T, out_poly.P, out_poly.H, K.power_utility.rate],
        [out_isen.T, out_isen.P, out_isen.H, K_isen.power_utility.rate],
        rtol=1e-3,
    )
    # eta=0.7, hundseid method
    eta=0.7
    feed = bst.Stream(H2=1, T=25 + 273.15, P=20e5, phase='g')
    K = bst.units.PolytropicCompressor(ins=feed, P=P, eta=eta, method="hundseid", n_steps=200)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, 958.0835733924658, P],
        rtol=1e-3,
    )
    # check compressor design
    assert K.design_results["Type"] == "Reciprocating"
    assert K.design_results["Driver"] == "Electric motor"
    assert len(K.heat_utilities) == 0 and K.net_duty == 0.
    expected_power = 6.484199755124557
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Compressors in parallel']],
        [expected_power, 1],
        rtol=1e-3,
    )
    pass


def test_multistage_hydrogen_compressor_simple():
    thermo = bst.Thermo([bst.Chemical('H2', eos=PR)])
    thermo.chemicals.H2.V.g.method_P = 'IDEAL'
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream("feed", H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    # test simple setup
    n_stages = 5
    pr = (P / feed.P) ** (1 / n_stages)
    K = bst.units.MultistageCompressor(ins=feed, outs=('outlet'), pr=pr, n_stages=n_stages, eta=0.7)
    K.simulate()
    # check inlet and outlet
    assert K.ins[0].ID == "feed"
    assert K.compressors[0].ins[0] == K.ins[0]
    assert K.outs[0].ID == "outlet"
    assert K.hxs[-1].outs[0] == K.outs[0]
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, feed.T, P],
        rtol=1e-3,
    )
    # check compressor design
    assert K.design_results["Type"] == "Multistage compressor"
    # assert K.design_results["Driver"] == "Electric motor"
    assert_allclose(
        [K.design_results["Area"]],
        [1.7906277579485619],
        rtol=1e-3,
    )
    # check heat utilities
    heat_utilities = bst.HeatUtility.sum_by_agent(K.heat_utilities)
    assert heat_utilities[1].ID == 'chilled_water'
    assert_allclose(
        [len(heat_utilities), heat_utilities[1].duty, heat_utilities[1].flow, heat_utilities[1].cost],
        [2, -11360.26358696248, 7.529257798470469, 0.056801317934812405],
        rtol=1e-3,
    )
    # check power utility
    assert_allclose(
        [K.power_utility.consumption, K.power_utility.production, K.power_utility.rate, K.power_utility.cost],
        [4.669790810710772, 3.86967926182525, 0.8001115488855222, 0.06256872312284784],
        rtol=1e-3,
    )
    pass


def test_multistage_hydrogen_compressor_advanced():
    thermo = bst.Thermo([bst.Chemical('H2', eos=PR)])
    thermo.chemicals.H2.V.g.method_P = 'IDEAL'
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream("feed", H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    # test advanced setup
    Ps = [60e5, 90e5, 200e5, 300e5, 350e5]
    Ts = [450, 400, 350, 300, feed.T]
    ks = [
        bst.units.IsentropicCompressor(P=Ps[0], eta=0.6),
        bst.units.PolytropicCompressor(P=Ps[1], eta=0.7),
        bst.units.IsentropicCompressor(P=Ps[2], eta=0.75),
        bst.units.PolytropicCompressor(P=Ps[3], eta=0.8),
        bst.units.IsentropicCompressor(P=Ps[4], eta=0.85),
    ]
    hxs = [bst.units.HXutility(T=T) for T in Ts]
    K = bst.units.MultistageCompressor(ins=feed, outs=('outlet'), compressors=ks, hxs=hxs)
    K.simulate()
    # check inlet and outlet
    assert K.ins[0].ID == "feed"
    assert K.compressors[0].ins[0] == K.ins[0]
    assert K.outs[0].ID == "outlet"
    assert K.hxs[-1].outs[0] == K.outs[0]
    # check hx outlet states
    for hx, P, T in zip(K.hxs, Ps, Ts):
        out = hx.outs[0]
        assert_allclose(
            [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
            [1.0, 0.0, T, P],
            rtol=1e-3,
        )
    # check compressor design
    assert K.design_results["Type"] == "Multistage compressor"
    # assert K.design_results["Driver"] == "Electric motor"
    assert_allclose(
        [K.design_results["Area"],],
        [1.1008174238336854],
        rtol=1e-3,
    )
    # check heat utilities
    heat_utilities = bst.HeatUtility.sum_by_agent(K.heat_utilities)
    assert len(heat_utilities) == 3
    assert heat_utilities[2].ID == 'chilled_water'
    assert heat_utilities[1].ID == 'cooling_water'
    assert heat_utilities[0].ID == 'high_pressure_steam'
    assert_allclose(
        [heat_utilities[2].duty, heat_utilities[2].flow, heat_utilities[2].cost],
        [-3694.721706001324, 2.4487558765814454, 0.018473608530006624],
        rtol=1e-3,
    )
    assert_allclose(
        [heat_utilities[1].duty, heat_utilities[1].flow, heat_utilities[1].cost],
        [-10377.599222102164, 7.087204175252751, 0.003457492556897055],
        rtol=1e-3,
    )
    # check power utility
    assert_allclose(
        [K.power_utility.consumption, K.power_utility.production, K.power_utility.rate, K.power_utility.cost],
        [6.021683407067501, 5.848476544603433, 0.1732068624640677, 0.013544776644690094],
        rtol=1e-3,
    )

def test_compressor_design():
    bst.settings.set_thermo(["H2"], cache=True)
    bst.settings.chemicals.H2.V.g.method_P = 'IDEAL'
    feed = bst.Stream(H2=1, T=298.15, P=1e5, phase='g')
    K = bst.units.IsothermalCompressor(ins=feed, P=350e5)
    
    # Test default compressor types across pressures:
    psig_to_pascal = lambda psig: (psig + 14.6959) * 101325 / 14.6959
    def compressor_type_at_psig(psig):
        K.P = psig_to_pascal(psig)
        K.simulate()
        return K.design_results["Type"]
    
    # Pressures [psig] in the same order of magnitude as upper bound for each
    # compressor type.
    pressures = [1e2, 1e3, 1e4] 
    compressor_types = ['Screw', 'Centrifugal', 'Reciprocating']
    assert [compressor_type_at_psig(i) for i in pressures] == compressor_types
    
    # Test compressors in parallel:
    # Actual cubic feet per minute in the same order of magnitude as upper
    # bound for each compressory type.
    acfms = [1.5e4, 1e5, 5e3]
    
    def get_compressors_in_parallel(psig, acfm):
        K.ins[0].set_total_flow(acfm, 'cfm')
        K.P = psig_to_pascal(psig)
        K.simulate()
        return K.design_results['Compressors in parallel']
    
    compressors_in_parallel = [1 , 2]
    for psig, acfm in zip(pressures, acfms):
        assert compressors_in_parallel == [get_compressors_in_parallel(psig, i * acfm) for i in compressors_in_parallel]
        
    # Test user overwrite:
    K.P = psig_to_pascal(1e3) 
    K.compressor_type = "Centrifugal"
    K.driver = 'Electric motor'
    K.driver_efficiency = 0.9
    K.simulate()
    assert K.design_results['Type'] == 'Centrifugal'
    assert K.design_results['Driver'] == 'Electric motor'
    assert K.design_results['Driver efficiency'] == 0.9
    
    # Test capital cost correlations and factors
    def installed_equipment_cost_at_psig(psig):
        K.P = psig_to_pascal(psig)
        K.simulate()
        return K.installed_cost
    
    K.driver = K.driver_efficiency = K.compressor_type = 'Default'
    assert_allclose([installed_equipment_cost_at_psig(i) for i in pressures],
                    [ 2267264.256152,  5179043.514644, 10479970.002024],
                    rtol=1e-3)

def test_multistage_setup_does_not_recreate_subcomponents():
    bst.settings.set_thermo(["H2"])
    bst.settings.chemicals.H2.V.g.method_P = 'IDEAL'
    feed = bst.Stream("feed", H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    n_stages = 5
    pr = (P / feed.P) ** (1 / n_stages)
    K = bst.units.MultistageCompressor(ins=feed, outs=('outlet'), pr=pr, n_stages=n_stages, eta=0.7)
    K._setup()
    a = K.compressors[0]
    K._setup()
    b = K.compressors[0]
    assert a==b
    
def test_multistage_setup_updates_after_changing_specifications():
    bst.settings.set_thermo(["H2"])
    bst.settings.chemicals.H2.V.g.method_P = 'IDEAL'
    feed = bst.Stream("feed", H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    n_stages = 5
    pr = (P / feed.P) ** (1 / n_stages)
    K = bst.units.MultistageCompressor(ins=feed, outs=('outlet'), pr=pr, n_stages=n_stages, eta=0.7)
    K._setup()
    units_5 = (K.compressors, K.hxs)
    for i in units_5: assert len(i) == 5
    K.n_stages = 6
    K._setup()
    units_6 = (K.compressors, K.hxs)
    for i in units_6: assert len(i) == 6


if __name__ == '__main__':
    test_compressor_design()
    test_isentropic_hydrogen_compressor()
    test_isentropic_two_phase_steam_compressor()
    test_isothermal_hydrogen_compressor()
    test_polytropic_hydrogen_compressor()
    test_multistage_hydrogen_compressor_simple()
    test_multistage_hydrogen_compressor_advanced()
    test_multistage_setup_does_not_recreate_subcomponents()
    test_multistage_setup_updates_after_changing_specifications()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from thermosteam._graphics import mixer_graphics
import biosteam as bst

__all__ = ('BlowdownMixer',)

class BlowdownMixer(bst.Facility, bst.Mixer):
    network_priority = 2
    _graphics = mixer_graphics
    ticket_name = 'BDM'